from django import apps
from django.utils.translation import gettext_lazy as _


class TreasuryConfig(apps.AppConfig):
    name = "erp_framework.erp.treasury"
    verbose_name = _("Treasury")

    def ready(self):
        super(TreasuryConfig, self).ready()



from __future__ import unicode_literals

import logging
from collections import OrderedDict
from collections.abc import Iterable

from django.apps import apps
from django.db.models import Max
from django.template.defaultfilters import capfirst
from django.urls import reverse, NoReverseMatch

logger = logging.getLogger(__name__)

get_model = apps.get_model


def get_obj_from_list(lst, key, val):
    return get_from_list(False, lst, key, val)


def get_from_list(by_getattr, lst, attr_name, val):
    """
    Gets an object from the supplied list by `getattr` or by `__item__`
    :param by_getattr: Use getattr, if False we use __item__ (dict access)
    :param lst: list of the objects
    :param attr_name: Or Key
    :param val: value we search for
    :return: The Item if found else return False
    """

    def get_by_key(_dict, _key):
        return _dict[_key]

    if by_getattr:
        lookup = getattr
    else:
        lookup = get_by_key
    for l in lst:
        if lookup(l, attr_name) == val:
            return l
    return False


class RaPermissionWidgetExclude(object):
    def __call__(self, model):
        from erp_framework.base.models import TransactionItemModel, BaseReportModel

        if (
            TransactionItemModel in model.__mro__
            or BaseReportModel in model.__mro__
            or model._meta.managed is False
        ):
            return True
        if model._meta.auto_created:
            return True
        return False


def order_apps(app_list):
    """Called in admin/base_site.html template override and applies custom ordering of
    apps/models defined by settings.ADMIN_REORDER
    """
    from . import app_settings

    # sort key function - use index of item in order if exists, otherwise item
    sort = (
        lambda order, item: (order.index(item), "")
        if item in order
        else (len(order), item)
    )

    # sort the app list
    order = OrderedDict(app_settings.ADMIN_REORDER)
    app_list.sort(
        key=lambda app: sort(order.keys(), app["app_url"].strip("/").split("/")[-1])
    )
    for i, app in enumerate(app_list):
        # sort the model list for each app
        app_name = app["app_url"].strip("/").split("/")[-1]
        model_order = [m.lower() for m in order.get(app_name, [])]
        app_list[i]["models"].sort(
            key=lambda model: sort(
                model_order, model.get("admin_url", "").strip("/").split("/")[-1]
            )
        )
    return app_list


def admin_get_app_list(request, admin_site):
    """
    :param request: Copied from AdminSite.index() djagno v1.8
    :param admin_site:
    :return:
    """
    from erp_framework.base.app_settings import RA_MENU_HIDE_MODELS, RA_ADMIN_SITE_NAME

    app_dict = {}
    for model, model_admin in admin_site._registry.items():
        app_label = model._meta.app_label
        has_module_perms = model_admin.has_module_permission(request)

        is_model_hidden = (
            "%s_%s" % (app_label, model.__name__.lower()) in RA_MENU_HIDE_MODELS
        )

        if has_module_perms and not is_model_hidden:
            perms = model_admin.get_model_perms(request)

            # Check whether user has any perm for this module.
            # If so, add the module to the model_list.
            if True in perms.values():
                info = (RA_ADMIN_SITE_NAME, app_label, model._meta.model_name)
                model_dict = {
                    "name": capfirst(model._meta.verbose_name_plural),
                    "object_name": model._meta.object_name,
                    "perms": perms,
                    "model_class": model,
                }
                if (
                    perms.get("view", False)
                    or perms.get("change", False)
                    or perms.get("add", False)
                ):
                    try:
                        model_dict["admin_url"] = reverse(
                            "%s:%s_%s_changelist" % info, current_app=admin_site.name
                        )
                    except NoReverseMatch:
                        pass
                if perms.get("add", False):
                    try:
                        model_dict["add_url"] = reverse(
                            "%s:%s_%s_add" % info, current_app=admin_site.name
                        )
                    except NoReverseMatch:
                        pass
                if app_label in app_dict:
                    app_dict[app_label]["models"].append(model_dict)
                else:
                    app_dict[app_label] = {
                        "name": apps.get_app_config(app_label).verbose_name,
                        "app_label": app_label,
                        "app_url": reverse(
                            "%s:app_list" % RA_ADMIN_SITE_NAME,
                            kwargs={"app_label": app_label},
                            current_app=admin_site.name,
                        ),
                        "has_module_perms": has_module_perms,
                        "models": [model_dict],
                    }

    # Sort the apps alphabetically.
    app_list = list(app_dict.values())
    app_list.sort(key=lambda x: x["name"].lower())

    # Sort the models alphabetically within each app.
    for app in app_list:
        app["models"].sort(key=lambda x: x["name"])

    return order_apps(app_list)


def dictsort(value, arg, desc=False):
    """
    Takes a list of dicts, returns that list sorted by the property given in
    the argument.
    """
    return sorted(value, key=lambda x: x[arg], reverse=desc)


#
# def get_ra_relevant_content_types():
#     """
#     This method scans the content type and show only what relevant based on the the exclude function supplied to
#     tabular_permissions
#     #todo make a separate function that might fall back to tabular_permissions
#     :return:
#     """
#     from django.contrib.contenttypes.models import ContentType
#     from erp_framework.base.models import BaseReportModel
#     from tabular_permissions.app_settings import EXCLUDE_FUNCTION
#
#     relevant_ct = []
#     cs = ContentType.objects.all()
#     exclude_function = EXCLUDE_FUNCTION
#     for c in cs:
#         model = c.model_class()
#         if not model:
#             continue
#         if BaseReportModel in model.__mro__:
#             continue
#         if model:
#             if not exclude_function(model):
#                 relevant_ct.append((c.pk, str(model._meta.verbose_name_plural)))
#
#     return relevant_ct


def get_next_serial(model, slug_field="slug"):
    """
    Get the next serial to put in the slug based on the maximum slug found + 1
    :param model: the model to get the next serial for
    :return: a string
    """
    import time

    qs = model.objects.aggregate(Max(slug_field))
    max_slug = qs.get(f"{slug_field}__max", 0) or 0
    if type(max_slug) is str and not max_slug.isdigit():
        return str(time.time()).split(".")[0]
    return int(max_slug) + 1


def default_formfield_for_dbfield(model_admin, db_field, form_field, request, **kwargs):
    """
    A system wide hook to change how db_field are displayed as formfields
    :param model_admin: the ModelAdmin instance
    :param db_field: db_field
    :param form_field: the default form_field
    :param request: the current request

    :return: form_field to be used
    """
    return form_field


def flatten_list(items):
    for x in items:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            yield from flatten_list(x)
        else:
            yield x



from django import template
from django.template.loader import get_template
from django.utils.functional import Promise
from django.utils.safestring import mark_safe
from logging import getLogger

logger = getLogger(__name__)

from erp_framework.utils.permissions import has_report_permission_permission

register = template.Library()

# def is_time_series(report, obj_instance, pk_name):
#     return report.get('time_series_pattern', '')


from django.core.serializers import serialize
from django.db.models.query import QuerySet
import simplejson as json


def jsonify(object):
    def date_handler(obj):
        if hasattr(obj, "isoformat"):
            return obj.isoformat()
        elif isinstance(obj, Promise):
            return str(obj)

    if isinstance(object, QuerySet):
        return serialize("json", object)

    return mark_safe(json.dumps(object, use_decimal=True, default=date_handler))


register.filter("jsonify", jsonify)


@register.simple_tag(takes_context=True)
def can_print_report(context, report):
    return has_report_permission_permission(
        context["request"], report.get_base_model_name(), report.get_report_slug()
    )


@register.filter
def addstr(arg1, arg2):
    """concatenate arg1 & arg2"""
    return str(arg1) + str(arg2)


@register.simple_tag
def get_html_panel(report, template="", **kwargs):
    kwargs["report"] = report
    if not report:
        raise ValueError(
            "report argument is empty. Are you sure you're using the correct report name"
        )

    # No chart argument default to True if no charts in reports
    kwargs.setdefault("no_chart", not bool(report.chart_settings))

    template = get_template(template or "erp_framework/report_widget_template.html")
    return template.render(context=kwargs)



# from __future__ import unicode_literals
import logging

import csv
from django.http import HttpResponse
from django.shortcuts import render

from erp_framework.utils.views import re_time_series
from erp_framework.base import app_settings

logger = logging.getLogger(__name__)


def regroup_data(obj_list, header_report, key):
    values = {}
    for line in obj_list:
        key_value = line[key]
        if key_value not in values:
            values[key_value] = []
        values[key_value].append(line)
    return values


# def is_time_series(name):
#     is_time_field = re_time_series.findall(name)
#     return is_time_field or False

#
# def is_partial_text_in_list(text, lst):
#     check = False
#     for item in lst:
#         if text in item:
#             check = True
#             break
#     return check
#


class HTMLPrintingClass(object):
    def __init__(
        self,
        request,
        form_settings,
        response,
        header_report=None,
        header_group=None,
        report_view=None,
        print_settings=None,
    ):
        self.reverse_when_bidi = True
        self.request = request
        self.response = response
        self.form_settings = form_settings
        self.get_group = True
        self.header_report = header_report
        self.group_by = form_settings.get("group_by", "")
        self.header_group = header_group
        self.report_view = report_view
        self.print_settings = print_settings
        self.report_title = (
            self.report_view.get_report_title()
            if getattr(report_view, "get_report_title", False)
            else ""
        )
        self.template_name = f"erp_framework/print_base.html"

    def get_response(self, template_name=None, extra_context=None):
        template_name = template_name or self.template_name
        extra_context = extra_context or {}
        context = {"response": self.response, "is_print": True}
        context.update(extra_context)
        return render(self.request, template_name, context)


class ExportToCSV(object):
    def get_filename(self):
        return (
            self.report_view.get_report_title()
            if getattr(self.report_view, "get_report_title", False)
            else ""
        )

    def get_response(self, template_name=None, extra_context=None):
        response = HttpResponse(content_type="text/csv")
        response["Content-Disposition"] = (
            'attachment; filename="%s.csv"' % self.get_filename()
        )

        writer = csv.writer(response)
        # for col_name in response['data']:
        # writer.writerow()
        col, col_names = self.get_columns(extra_context)
        writer.writerow(col_names)
        for line in self.response["data"]:
            export_line = []
            for c in col:
                export_line.append(line[c])

            writer.writerow(export_line)

        return response

    def get_columns(self, extra_context=None):
        cols, col_names = list(
            zip(*[(x["name"], x["verbose_name"]) for x in self.response["columns"]])
        )
        # col_names = self.response['column_names']
        result_cols = []
        result_col_names = []
        for i, c in enumerate(cols):
            if "_id" not in c and c != "_control_":
                result_cols.append(c)
                result_col_names.append(col_names[i])
        return result_cols, result_col_names

    def __init__(
        self,
        request,
        form_settings,
        response,
        header_report=None,
        header_group=None,
        report_view=None,
        print_settings=None,
    ):
        self.reverse_when_bidi = True
        self.request = request
        self.response = response
        self.form_settings = form_settings
        self.get_group = True
        self.header_report = header_report
        self.group_by = form_settings.get("group_by", "")
        self.header_group = header_group
        self.report_view = report_view
        self.print_settings = print_settings
        self.report_title = (
            self.report_view.get_report_title()
            if getattr(report_view, "get_report_title", False)
            else ""
        )



# Generated by Django 2.2.12 on 2020-07-11 12:53

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('activity', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='myactivity',
            options={'verbose_name': 'Activity', 'verbose_name_plural': 'Activities'},
        ),
    ]



from __future__ import unicode_literals

from django.contrib.auth.forms import UserChangeForm
from django.utils.translation import gettext_lazy as _


class RaUserChangeForm(UserChangeForm):
    def __init__(self, *args, **kwargs):
        super(RaUserChangeForm, self).__init__(*args, **kwargs)
        self.fields['password'].help_text = '<a class="reset-password" href=\"../password/\">%s</a>' % _(
            'Reset password here')


# class CtCheckForm(forms.Form):
#     content_type = forms.ChoiceField(choices=(), widget=forms.Select(), label=_('content type'))
#     object_slug = forms.CharField(max_length=300, label=_('slug'))
#
#     def __init__(self, *args, **kwargs):
#         self.base_fields['content_type'] = forms.ChoiceField(choices=get_ra_relevant_content_types(),
#                                                              widget=forms.Select())
#         super(CtCheckForm, self).__init__(*args, **kwargs)
#
#     def clean_content_type(self):
#         pk = self.cleaned_data['content_type']
#         return ContentType.objects.get_for_id(pk)


# class CtReverseCheckForm(forms.Form):
#     content_type = forms.ChoiceField(choices=(), widget=forms.Select(), label=_('content type'))
#     object_id = forms.IntegerField()
#
#     def __init__(self, *args, **kwargs):
#         self.base_fields['content_type'] = forms.ChoiceField(choices=get_ra_relevant_content_types(),
#                                                              widget=forms.Select())
#         super(CtReverseCheckForm, self).__init__(*args, **kwargs)
#
#     def clean_content_type(self):
#         pk = self.cleaned_data['content_type']
#         return ContentType.objects.get_for_id(pk)



from inspect import getfullargspec

from django.template import Library
from django.template.library import InclusionNode, parse_bits
from django.template.loader import get_template

register = Library()
from django.contrib.admin.templatetags.admin_list import (
    result_list,
    pagination,
    date_hierarchy,
    search_form,
    admin_actions,
)
from django.contrib.admin.templatetags.admin_modify import submit_row

from erp_framework.base import app_settings


class InclusionAdminNode(InclusionNode):
    """
    Template tag that allows its template to be overridden per model, per app,
    or globally.
    """

    def __init__(self, parser, token, func, template_name, takes_context=True):
        self.template_name = template_name
        params, varargs, varkw, defaults, kwonly, kwonly_defaults, _ = getfullargspec(
            func
        )
        bits = token.split_contents()
        args, kwargs = parse_bits(
            parser,
            bits[1:],
            params,
            varargs,
            varkw,
            defaults,
            kwonly,
            kwonly_defaults,
            takes_context,
            bits[0],
        )
        super().__init__(func, takes_context, args, kwargs, filename=None)

    def render(self, context):
        opts = context["opts"]
        app_label = opts.app_label.lower()
        object_name = opts.object_name.lower()
        # Load template for this render call. (Setting self.filename isn't
        # thread-safe.)
        context.render_context[self] = context.template.engine.select_template(
            [
                "%s/%s/%s" % (app_label, object_name, self.template_name),
                "%s/%s" % (app_label, self.template_name),
                "%s" % (self.template_name,),
            ]
        )
        return super().render(context)


@register.tag(name="ra_result_list")
def result_list_tag(parser, token):
    return InclusionAdminNode(
        parser,
        token,
        func=result_list,
        template_name=f"{app_settings.RA_THEME}/change_list_results.html",
        takes_context=False,
    )


@register.tag(name="ra_pagination")
def pagination_tag(parser, token):
    return InclusionAdminNode(
        parser,
        token,
        func=pagination,
        template_name=f"{app_settings.RA_THEME}/pagination.html",
        takes_context=False,
    )


@register.tag(name="ra_date_hierarchy")
def date_hierarchy_tag(parser, token):
    return InclusionAdminNode(
        parser,
        token,
        func=date_hierarchy,
        template_name=f"{app_settings.RA_THEME}/date_hierarchy.html",
        takes_context=False,
    )


@register.tag(name="ra_search_form")
def search_form_tag(parser, token):
    return InclusionAdminNode(
        parser,
        token,
        func=search_form,
        template_name=f"{app_settings.RA_THEME}/search_form.html",
        takes_context=False,
    )


@register.tag(name="ra_admin_actions")
def admin_actions_tag(parser, token):
    return InclusionAdminNode(
        parser,
        token,
        func=admin_actions,
        template_name=f"{app_settings.RA_THEME}/actions.html",
    )


@register.tag(name="ra_change_list_object_tools")
def change_list_object_tools_tag(parser, token):
    """Display the row of change list object tools."""
    return InclusionAdminNode(
        parser,
        token,
        func=lambda context: context,
        template_name=f"{app_settings.RA_THEME}/change_list_object_tools.html",
    )


@register.tag(name="ra_change_form_object_tools")
def change_form_object_tools_tag(parser, token):
    """Display the row of change form object tools."""
    return InclusionAdminNode(
        parser,
        token,
        func=lambda context: context,
        template_name=f"{app_settings.RA_THEME}/change_form_object_tools.html",
    )


@register.simple_tag
def ra_admin_list_filter(cl, spec):
    tpl = get_template(f"{app_settings.RA_THEME}/filter.html")
    return tpl.render(
        {
            "title": spec.title,
            "choices": list(spec.choices(cl)),
            "spec": spec,
        }
    )


@register.tag(name="ra_submit_row")
def submit_row_tag(parser, token):
    return InclusionAdminNode(
        parser,
        token,
        func=submit_row,
        template_name=f"{app_settings.RA_THEME}/submit_line.html",
    )



from __future__ import unicode_literals

from erp_framework.reporting.registry import report_registry

from django.apps import apps as global_apps
from django.contrib.contenttypes.management import create_contenttypes
from django.db import DEFAULT_DB_ALIAS, router


def _get_all_permissions(report):
    """
    Return (codename, name) for all permissions in the given opts.
    """
    return [*_get_builtin_permissions(report)]


def _get_builtin_permissions(report):
    """
    Return (codename, name) for all autogenerated permissions.
    By default, this is ('add', 'change', 'delete', 'view')
    """
    default_permissions = ["print", "view"]
    perms = []
    for action in default_permissions:
        permission_name = "%s_%s" % (action, report.get_report_slug())
        perms.append((permission_name, "Can %s %s" % (action, report.report_title)))
    return perms


def create_report_permissions(
    app_config,
    verbosity=2,
    interactive=True,
    using=DEFAULT_DB_ALIAS,
    apps=global_apps,
    **kwargs
):
    """
    Copied from django create permissions for model, edited to create view and print for reports
    Reports permissions are created in the database similar to django's

    app.print_clientbalances (ClientBalances is the report slug (defaulting to report class name), Content Type is the
    base model content_type

    :param app_config:
    :param verbosity:
    :param interactive:
    :param using:
    :param apps:
    :param kwargs:
    :return:
    """

    if not app_config.models_module:
        return

    # Ensure that contenttypes are created for this app. Needed if
    # 'django.contrib.auth' is in INSTALLED_APPS before
    # 'django.contrib.contenttypes'.
    create_contenttypes(
        app_config,
        verbosity=verbosity,
        interactive=interactive,
        using=using,
        apps=apps,
        **kwargs
    )

    app_label = app_config.label
    try:
        app_config = apps.get_app_config(app_label)
        ContentType = apps.get_model("contenttypes", "ContentType")
        Permission = apps.get_model("auth", "Permission")
    except LookupError:
        return

    if not router.allow_migrate_model(using, Permission):
        return

    # This will hold the permissions we're looking for as
    # (content_type, (codename, name))
    all_reports = report_registry.get_all_reports()
    searched_perms = []
    # The codenames and ctypes that should exist.
    ctypes = set()
    for report in all_reports:
        # for klass in report_registry.get_base_models():
        # Force looking up the content types in the current database
        # before creating foreign keys to them.
        klass = report.base_model
        ctype = ContentType.objects.db_manager(using).get_for_model(
            klass, for_concrete_model=False
        )

        ctypes.add(ctype)
        for perm in _get_all_permissions(report):
            searched_perms.append((ctype, perm))

    # Find all the Permissions that have a content_type for a model we're
    # looking for.  We don't need to check for codenames since we already have
    # a list of the ones we're going to create.
    all_perms = set(
        Permission.objects.using(using)
        .filter(
            content_type__in=ctypes,
        )
        .values_list("content_type", "codename")
    )

    perms = [
        Permission(codename=codename, name=name, content_type=ct)
        for ct, (codename, name) in searched_perms
        if (ct.pk, codename) not in all_perms
    ]
    Permission.objects.using(using).bulk_create(perms)
    if verbosity >= 2:
        for perm in perms:
            print("Adding permission '%s'" % perm)


def has_report_permission_permission(request, namespace, slug):
    user = request.user
    if user.is_superuser:
        return True
    else:
        return "%s.%s_print" % (namespace, slug) in request.session.get(
            "user_reports", ""
        )



import datetime
import logging

from crequest.middleware import CrequestMiddleware
from django.contrib.auth import get_user_model
from django.db import models
from django.urls import reverse, NoReverseMatch
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from . import app_settings, registry

logger = logging.getLogger(__name__)
User = get_user_model()


class DiffingMixin(object):
    def __init__(self, *args, **kwargs):
        super(DiffingMixin, self).__init__(*args, **kwargs)
        self._original_state = dict(self.__dict__)
        self.pre_current_state = None

    def save(self, *args, **kwargs):
        state = dict(self.__dict__)
        del state["_original_state"]
        super(DiffingMixin, self).save(*args, **kwargs)
        self.pre_current_state = self._original_state
        self._original_state = state

    def is_dirty(self):
        missing = object()
        for key, value in self._original_state.items():
            try:
                if value != self.__dict__.get(key, missing):
                    return True
            except TypeError as e:
                o = self.__dict__.get(key, missing)
                if type(value) == datetime.datetime and type(o) == datetime.datetime:
                    # Make both unaware
                    o = o.replace(tzinfo=None)
                    value = value.replace(tzinfo=None)
                    if value != o:
                        return True

                else:
                    raise e
        return False

    def changed_columns(self):
        missing = object()
        result = {}
        for key, value in self._original_state.items():
            try:
                if value != self.__dict__.get(key, missing):
                    result[key] = {"old": value, "new": self.__dict__.get(key, missing)}
            except TypeError:
                result[key] = {"old": value, "new": self.__dict__.get(key, missing)}
        return result


class RAModel(DiffingMixin, models.Model):
    owner = models.ForeignKey(
        User,
        related_name="%(app_label)s_%(class)s_related",
        verbose_name=_("owner"),
        on_delete=models.CASCADE,
    )
    creation_date = models.DateTimeField(_("creation date and time"), default=now)
    lastmod = models.DateTimeField(_("last modification"), db_index=True)
    lastmod_user = models.ForeignKey(
        User,
        related_name="%(app_label)s_%(class)s_lastmod_related",
        verbose_name=_("last modification by"),
        on_delete=models.CASCADE,
    )

    class Meta:
        abstract = True


class ERPMixin:
    @classmethod
    def get_class_name(cls):
        """
        return the class name, usable when a ra model is mimicking (ie:proxying)
        another model.
        This method is used is get_doc_type_* functions,
        This method is made to avoid to repeat registered type to make adjustments
        """
        return cls.__name__

    @classmethod
    def get_model_name(cls):
        """
        A convenience method to get the base model name, needed in templates
        :return:
        """

        return cls._meta.model_name.lower()

    @classmethod
    def get_verbose_name_plural(cls):
        """
        A convenience method to get the base model verbose name, needed in templates
        :return:
        """
        return cls._meta.verbose_name_plural


class EntityModel(ERPMixin, RAModel):
    """
    The Main base for Ra `static` models
    Example: Client , Expense etc..
    """

    slug = models.SlugField(
        _("Identifier slug"),
        help_text=_("For fast recall"),
        max_length=50,
        unique=True,
        db_index=True,
        blank=True,
    )
    name = models.CharField(_("Name"), max_length=255, unique=True, db_index=True)
    notes = models.TextField(_("Notes"), null=True, blank=True)

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):
        super(EntityModel, self).__init__(*args, **kwargs)
        # self.reporting_model = None
        if not getattr(self, "pk_name", False):
            self.pk_name = None

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        model_name = self._meta.model_name.lower()
        try:
            url = reverse(
                "%s:%s_%s_view"
                % (app_settings.RA_ADMIN_SITE_NAME, self._meta.app_label, model_name),
                args=(self.pk,),
            )
        except NoReverseMatch:
            url = reverse(
                "%s:%s_%s_change"
                % (
                    app_settings.RA_ADMIN_SITE_NAME,
                    self._meta.app_label,
                    self.get_class_name().lower(),
                ),
                args=(self.pk,),
            )
        return url

    def get_next_slug(self, suggestion=None):
        """
        Get the next slug
        If it's a new instance and the slug is not provided, we try and attempt a serial over the already added slugs
        in relation to the model
        :return:
        """
        from .helpers import get_next_serial

        return get_next_serial(self.__class__)  # repr(time.time()).replace('.', '')

    def save(self, *args, **kwargs):
        if self.pk is None:
            if not self.slug:
                self.slug = self.get_next_slug(self.name)
                # print(self.slug)
            if not self.owner_id:
                try:
                    self.owner = self.lastmod_user
                except:
                    self.owner_id = self.lastmod_user_id
                    logger.info(
                        "lastmod_user_id is used instead of lastmod_user object"
                    )

        self.lastmod = now()

        super(EntityModel, self).save(*args, **kwargs)

    @classmethod
    def _get_doc_type_plus_list(cls):
        """

        Returns List of Identified doctype that a plus effect on the entity
        """
        return ["fb"] + registry.get_model_doc_type_map(cls.get_class_name()).get(
            "plus_list", []
        )

    @classmethod
    def _get_doc_type_minus_list(cls):
        """Returns List of Identified doctype that a minus effect on the entity"""

        return registry.get_model_doc_type_map(cls.get_class_name()).get(
            "minus_list", []
        )

    @classmethod
    def get_doc_type_neuter_list(cls):
        """Returns List of Identified doctype that have a neuttral effect on the entity"""

        return []

    #
    # @classmethod
    # def get_doc_type_full_map(cls):
    #     from .registry import model_doc_type_full_map
    #
    #     doc_types_unfiltered = model_doc_type_full_map.get(cls.get_class_name(), [])
    #     doc_typed_filtered = []
    #     for type in doc_types_unfiltered:
    #         if not type.get('hidden', False):
    #             doc_typed_filtered.append(type)
    #
    #     return doc_typed_filtered

    # @classmethod
    # def get_doc_types(cls):
    #     """
    #     Return a list of the doc_types supported by the current model , Must implemented when needed by children
    #     @return:
    #     """
    #     return cls._get_doc_type_plus_list() + cls._get_doc_type_minus_list() + cls.get_doc_type_neuter_list()

    def get_pk_name(self):
        """
        This is used to get the full name of the primary key,
        a bit hackish but is important for reports.
        :return:
        """
        if self.pk_name:
            return self.pk_name
        else:
            # return self._meta.pk.column #not now
            return "%s_id" % self.__class__.__name__.lower()

    def get_title(self):
        """
        A helper function to get a custom name of the instance if needed
        :return:
        """
        return self.name

    @classmethod
    def get_report_list_url(cls):
        """
        Return the url for the report list for this model
        :return: a string url
        """

        return reverse(
            "%s:report_list" % app_settings.RA_ADMIN_SITE_NAME,
            args=(cls.get_class_name().lower(),),
        )

    @classmethod
    def get_redirect_url_prefix(cls):
        """
        Get the url for the change list of this model
        :return: a string url
        """
        return reverse(
            "%s:%s_%s_changelist"
            % (
                app_settings.RA_ADMIN_SITE_NAME,
                cls._meta.app_label,
                cls.get_class_name().lower(),
            )
        )


# class BasePersonInfo(BaseInfo):
#     address = models.CharField(_('address'), max_length=260, null=True, blank=True)
#     telephone = models.CharField(_('telephone'), max_length=130, null=True, blank=True)
#     email = models.EmailField(_('email'), null=True, blank=True)
#
#     class Meta:
#         abstract = True
#         # swappable = swapper.swappable_setting('ra', 'BasePersonInfo')


class TransactionModel(EntityModel):
    name = None

    slug = models.SlugField(
        _("Slug"), max_length=50, db_index=True, validators=[], blank=True
    )
    date = models.DateTimeField(_("Date"), db_index=True)
    type = models.CharField(max_length=30, db_index=True)
    notes = models.TextField(_("Notes"), null=True, blank=True)
    value = models.DecimalField(_("Value"), max_digits=19, decimal_places=2, default=0)

    owner = models.ForeignKey(
        User,
        related_name="%(app_label)s_%(class)s_related",
        verbose_name=_("Created By"),
        on_delete=models.CASCADE,
    )
    creation_date = models.DateTimeField(_("Creation timestamp"), default=now)
    lastmod = models.DateTimeField(_("Last modification"), db_index=True)
    lastmod_user = models.ForeignKey(
        User,
        related_name="%(app_label)s_%(class)s_lastmod_related",
        verbose_name=_("Last modification by"),
        on_delete=models.CASCADE,
    )

    @classmethod
    def get_doc_type(cls):
        """
        Return the type
        :return:
        """
        return cls.__name__.lower()
        # raise NotImplementedError(
        #     f'Class {cls} dont have a get_doc_type override. Each Transaction should define a *type*')

    def __str__(self):
        return "%s-%s" % (self._meta.verbose_name, self.slug)

    def __repr__(self):
        return "<%s pk:%s slug:%s type:%s>" % (
            self.__class__.__name__,
            self.pk,
            self.slug,
            self.type,
        )

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        """
        Custom save, it assign the user  As owner and the last modifed
        it sets the type
        make sure that dlc_date has correct timezone ?!
        :param force_insert:
        :param force_update:
        :param using:
        :param update_fields:
        :return:
        """

        from erp_framework.base.helpers import get_next_serial

        request = CrequestMiddleware.get_request()
        self.type = self.get_doc_type()
        if not self.slug:
            self.slug = get_next_serial(self.__class__)

        # self.slug = slugify(self.slug)
        if not self.pk:
            if not self.lastmod_user_id:
                self.lastmod_user_id = request.user.pk
            if not self.owner_id:
                self.owner_id = request.user.pk
            self.date = self.date if self.date else now()
        self.lastmod = now()

        super(EntityModel, self).save(*args, **kwargs)

    #
    # @classmethod
    # def get_doc_type_verbose_name(cls, type):
    #     """
    #     Return the type verbose name , Must be implemented when needed by children
    #     @param type: the type field value
    #     @return: the description of the type
    #     Example: In: get_doc_type_verbose_name('1')
    #             Out: Purchase
    #     """
    #     # Example :
    #     # if type == '1': return _('purchase')
    #     raise NotImplemented()

    # def get_absolute_url(self):
    #     doc_types = registry.get_doc_type_settings()
    #     if self.type in doc_types:
    #         return '%sslug/%s/' % (doc_types[self.type]['redirect_url_prefix'], self.slug)
    #     else:
    #         return self.type

    @property
    def name(self):
        return self.date.strftime("%Y/%m/%d %H:%M")


class TransactionItemModel(TransactionModel):
    """
    Abstract model to identify a movement with a value
    """

    class Meta:
        abstract = True


class QuantitativeTransactionItemModel(TransactionItemModel):
    quantity = models.DecimalField(
        _("Quantity"), max_digits=19, decimal_places=2, default=0
    )
    price = models.DecimalField(_("Price"), max_digits=19, decimal_places=2, default=0)
    discount = models.DecimalField(
        _("Discount"), max_digits=19, decimal_places=2, default=0
    )

    def save(self, *args, **kwargs):
        self.value = self.quantity * self.price
        if self.discount:
            self.value -= self.value * self.discount / 100
        super(QuantitativeTransactionItemModel, self).save(*args, **kwargs)

    class Meta:
        abstract = True


class BaseReportModel(DiffingMixin, models.Model):
    slug = models.SlugField(
        _("Slug"), max_length=50, db_index=True, validators=[], blank=True
    )
    date = models.DateTimeField(_("Date"), db_index=True)
    type = models.CharField(max_length=30, db_index=True, verbose_name=_("Type"))
    notes = models.TextField(_("Notes"), null=True, blank=True)
    value = models.DecimalField(_("Value"), max_digits=19, decimal_places=2, default=0)

    owner = models.ForeignKey(
        User,
        related_name="%(app_label)s_%(class)s_related",
        verbose_name=_("Created By"),
        on_delete=models.DO_NOTHING,
    )
    creation_date = models.DateTimeField(_("Creation Timestamp"), default=now)
    lastmod = models.DateTimeField(_("Last modified"), db_index=True)
    lastmod_user = models.ForeignKey(
        User,
        related_name="%(app_label)s_%(class)s_lastmod_related",
        verbose_name=_("Last modification by"),
        on_delete=models.DO_NOTHING,
    )

    @classmethod
    def get_doc_type_plus_list(cls):
        """

        Returns List of Identified doctype that a plus effect on the entity
        """
        return []  # ['0','3' , 'client-cash-in','supplier-cash-in']

    @classmethod
    def get_doc_type_minus_list(cls):
        """Returns List of Identified doctype that a minus effect on the entity"""

        return []  # ['1', '2', 'client-cash-out', 'supplier-cash-out']

    # @classmethod
    # def get_doc_types(cls):
    #     """
    #     Return a list of the doc_types supported by the current model , Must implemented when needed by children
    #     @return:
    #     """
    #     return cls.get_doc_type_plus_list() + cls.get_doc_type_minus_list()

    class Meta:
        abstract = True


class QuanValueReport(BaseReportModel):
    quantity = models.DecimalField(
        _("quantity"), max_digits=19, decimal_places=2, default=0
    )
    price = models.DecimalField(_("price"), max_digits=19, decimal_places=2, default=0)
    discount = models.DecimalField(
        _("discount"), max_digits=19, decimal_places=2, default=0
    )

    @classmethod
    def get_doc_type_plus_list(cls):
        """

        Returns List of Identified doctype that a plus effect on the entity
        """
        return []  # ['0','3' , 'client-cash-in','supplier-cash-in']

    @classmethod
    def get_doc_type_minus_list(cls):
        """Returns List of Identified doctype that a minus effect on the entity"""

        return []  # ['1', '2', 'client-cash-out', 'supplier-cash-out']

    # @classmethod
    # def get_doc_types(cls):
    #     """
    #     Return a list of the doc_types supported by the current model , Must implemented when needed by children
    #     @return:
    #     """
    #     return cls.get_doc_type_plus_list() + cls.get_doc_type_minus_list()

    class Meta:
        abstract = True


class ProxyMovementManager(models.Manager):
    def get_queryset(self):
        return (
            super(ProxyMovementManager, self)
            .get_queryset()
            .filter(type=self.model.get_doc_type())
        )


class ProxyMovement(object):
    objects = ProxyMovementManager()

    @classmethod
    def get_doc_type(cls):
        """
        Get the doc-type of the row.
        This method is called internally during save to ensure that the record in the
        proxy model always have the right type
        :return: string (type)
        """
        return ""

    def __init__(self, *args, **kwargs):
        super(ProxyMovement, self).__init__(*args, **kwargs)
        self._doc_type = ""

    def save(self, *args, **kwargs):
        self.type = self.__class__.get_doc_type()
        super(ProxyMovement, self).save(*args, **kwargs)

    class Meta:
        proxy = True



from django.apps import apps
from django.contrib.admin.sites import NotRegistered
from django.urls import NoReverseMatch

# from .loading import get_baseinfo_model


class RaModelsRegistry(object):
    def __init__(self):
        super(RaModelsRegistry, self).__init__()
        self._registry = {}

    def populate(self):
        if self._registry:
            return

        models = apps.get_models()
        for m in models:
            self._registry[m._meta.model_name.lower()] = m

    def get(self, model_name):
        try:
            model = self._registry[model_name.lower()]
        except:
            options = self._registry.keys()
            raise NotRegistered(
                f'Model name {model_name} not identified. options are {options}. \n Are you sure that the model exsits and its '
                'app is added to INSTALLED_APPS ?')

        return model


_models = None
ra_model_registry = RaModelsRegistry()

#
# def get_ra_models():
#     # BaseInfo = get_baseinfo_model()
#     # from . models import CostCenter
#     global _models
#     if not _models:
#         model_list = apps.get_models()
#         ra_model_list = [x for x in model_list if BaseInfo in x.__mro__]
#         # ra_model_list = [x for x in model_list if BaseInfo in x.__mro__ or x is CostCenter]
#         _models = ra_model_list
#     return _models


def get_ra_model_by_name(model_name):
    ra_model_registry.populate()
    return ra_model_registry.get(model_name)


_doc_types = []
model_doc_type_map = {}
registry = ''

models_settings = {}
_doc_type_map = {}
model_doc_type_full_map = {}


def register_doc_type(*doc_types):
    for doc_type in doc_types:
        _doc_types.append(doc_type)

        _doc_type_map[doc_type['name']] = doc_type
        if 'plus_list' in doc_type:

            for plus_item in doc_type['plus_list']:
                if not plus_item in model_doc_type_full_map:
                    model_doc_type_full_map[plus_item] = []
                model_doc_type_full_map[plus_item].append(doc_type)

                if not plus_item in model_doc_type_map:
                    model_doc_type_map[plus_item] = {'plus_list': [], 'minus_list': []}
                model_doc_type_map[plus_item]['plus_list'].append(doc_type['name'])
        if 'minus_list' in doc_type:
            for plus_item in doc_type['minus_list']:
                if not plus_item in model_doc_type_map:
                    model_doc_type_map[plus_item] = {'plus_list': [], 'minus_list': []}
                model_doc_type_map[plus_item]['minus_list'].append(doc_type['name'])

                if not plus_item in model_doc_type_full_map:
                    model_doc_type_full_map[plus_item] = []
                model_doc_type_full_map[plus_item].append(doc_type)


# def get_doc_type_settings():
#     return _doc_type_map


def get_model_doc_type_map(model_name):
    return model_doc_type_map.get(model_name, {'plus_list': [], 'minus_list': []})

#
# def get_model_settings():
#     if not models_settings:
#         _fill_models_settings()
#     return models_settings
#
#
# def _fill_models_settings():
#     ra_models = [] # list(get_ra_models())
#     for model in ra_models:
#         try:
#             url = model.get_redirect_url_prefix()
#         except NoReverseMatch:
#             url = ''
#
#         models_settings[model.__name__.lower()] = {'model': model, 'redirect_url_prefix': url}



from __future__ import unicode_literals

from django.core.cache import cache
from django.conf import settings
from django.utils.translation import get_language

from erp_framework.base.helpers import admin_get_app_list


def get_each_context(request, admin_site):
    context = {}
    current_language = get_language() or ""
    cache_key = "app_list_%s_%s" % (request.user.pk, current_language)
    cache_val = cache.get(cache_key)
    if not cache_val or settings.DEBUG:
        cache_val = admin_get_app_list(request, admin_site)
        cache.set(cache_key, cache_val)
    context["app_list"] = cache_val
    context["admin_site"] = admin_site
    return context



from __future__ import unicode_literals

from django.template.loader import get_template

from erp_framework.base import app_settings


class RaSuitMenu(object):
    @classmethod
    def get_menu(cls, context, request, admin_site):
        from erp_framework.admin.templatetags.suit_menu import get_menu as suit_get_menu

        context["app_list"] = suit_get_menu(context, request)
        return get_template("ra/menu.html").render(context.flatten())



# coding=utf-8
import pytz
from django.conf import settings
from django.urls import get_callable
from django.utils.functional import lazy

# def get_first_of_this_year():
#     d = datetime.datetime.today()
#     return datetime.datetime(d.year, 1, 1, 0, 0, 0, 0, pytz.timezone(settings.TIME_ZONE))
#
#
# def get_end_of_this_year():
#     d = datetime.datetime.today()
#     return datetime.datetime(d.year + 1, 1, 1, 0, 0, 0, 0, pytz.timezone(settings.TIME_ZONE))


import datetime

ERP_FRAMEWORK_SETTING_DEFAULT = {
    "site_name": "ERP Framework System",
    "site_header": "ERP Framework System",
    "index_title": "ERP Framework Dashboard",
}

ERP_FRAMEWORK_SETTINGS = getattr(
    settings, "ERP_FRAMEWORK_SETTINGS", ERP_FRAMEWORK_SETTING_DEFAULT
)


"""
Documented
"""
# a function to control be dbfield on all instances, Saves you time to subclass ifonly you need to add a help text or something
RA_FORMFIELD_FOR_DBFIELD_FUNC = getattr(
    settings,
    "RA_FORMFIELD_FOR_DBFIELD_FUNC",
    "erp_framework.base.helpers.default_formfield_for_dbfield",
)
RA_FORMFIELD_FOR_DBFIELD_FUNC = get_callable(RA_FORMFIELD_FOR_DBFIELD_FUNC)
# ------

# Navigation class
RA_NAVIGATION_CLASS = getattr(
    settings, "RA_NAVIGATION_CLASS", "erp_framework.utils.navigation.RaSuitMenu"
)

"""
UnDocumented
"""

RA_ENABLE_ADMIN_DELETE_ALL = getattr(settings, "RA_ENABLE_ADMIN_DELETE_ALL", False)

# RA_DEFAULT_FROM_DATETIME = lazy(get_first_of_this_year, datetime.datetime)()
# RA_DEFAULT_TO_DATETIME = lazy(get_end_of_this_year, datetime.datetime)()
# # models


RA_ADMIN_SITE_NAME = getattr(settings, "RA_ADMIN_SITE_NAME", "erp_admin")

RA_REPORT_LIST_MAP = getattr(settings, "RA_REPORT_LIST_MAP", {})

RA_MENU_HIDE_MODELS = getattr(settings, "RA_MENU_HIDE_MODELS", [])
ADMIN_REORDER = getattr(settings, "ADMIN_REORDER", [])

RA_FLOATFORMAT_ARG = getattr(settings, "RA_FLOATFORMAT_ARG", "-2")

# Cache
# =====

RA_CACHE_REPORTS = getattr(settings, "RA_CACHE_REPORTS", True)
RA_CACHE_REPORTS_PER_USER = getattr(settings, "RA_CACHE_REPORTS_PER_USER", True)

# Style
RA_THEME = getattr(settings, "RA_THEME", "admin")

# Admin Looks
from django.utils.translation import gettext_lazy as _

RA_ADMIN_INDEX_TEMPLATE = getattr(
    settings, "RA_ADMIN_INDEX_PAGE", f"{RA_THEME}/index.html"
)
RA_ADMIN_APP_INDEX_TEMPLATE = getattr(
    settings, "RA_ADMIN_APP_INDEX_PAGE", f"{RA_THEME}/app_index.html"
)
RA_ADMIN_LOGIN_TEMPLATE = getattr(
    settings, "RA_ADMIN_LOGIN_TEMPLATE", f"{RA_THEME}/login.html"
)
RA_ADMIN_LOGGED_OUT_TEMPLATE = getattr(
    settings, "RA_ADMIN_LOGIN_TEMPLATE", f"{RA_THEME}/logged_out.html"
)

RA_ADMIN_SITE_CLASS = getattr(
    settings, "RA_ADMIN_SITE_CLASS", "erp_framework.admin.admin.RaAdminSite"
)

RA_ADMIN_SITE_TITLE = ERP_FRAMEWORK_SETTINGS.get("site_name", "ERP Framework System")

RA_ADMIN_SITE_HEADER = ERP_FRAMEWORK_SETTINGS.get("site_name", "ERP Framework System")

RA_ADMIN_INDEX_TITLE = ERP_FRAMEWORK_SETTINGS.get(
    "index_title", "ERP Framework Dashboard"
)



import re

eastern_western_map = {
    1776: 48,  # 0
    1777: 49,  # 1
    1778: 50,  # 2
    1779: 51,  # 3
    1780: 52,  # 4
    1781: 53,  # 5
    1782: 54,  # 6
    1783: 55,  # 7
    1784: 56,  # 8
    1785: 57,  # 9
    # another ord
    1632: 48,  # 0
    1633: 49,  # 1
    1634: 50,  # 2
    1635: 51,  # 3
    1636: 52,  # 4
    1637: 53,  # 5
    1638: 54,  # 6
    1639: 55,  # 7
    1640: 56,  # 8
    1641: 57,  # 9
}  # 9

re_time_series = re.compile("TS\d+")


def get_typed_reports_map(typed_reports, only_report_slug=None):
    """
    # todo revise
    :param typed_reports:
    :param only_report_slug:
    :return:
    """
    reports = typed_reports

    reports_map = {
        "slugs": [],
        "reports": [],
    }

    for report in reports:
        view = report
        if not only_report_slug or only_report_slug == view.get_report_slug():
            if True:  # user.has_perm(view_perm):
                reports_map["slugs"].append(view.get_report_slug())
                reports_map["reports"].append(view)
    return reports_map


def apply_order_to_typed_reports(lst, order_list):
    values = []
    unordered = list(lst)
    for o in order_list:
        for x in unordered:
            if x.get_report_slug() == o:
                values.append(x)
                unordered.remove(x)
    values += unordered
    return values


def get_typed_reports_for_templates(
    model_name, user=None, request=None, only_report_slug=None, load_func=None
):
    from erp_framework.reporting.registry import report_registry

    load_func = load_func or report_registry.get_report_classes_by_namespace
    reports = load_func(model_name)
    report_list = []
    user_reports = request.session.get("user_reports", [])
    for report in reports:
        view = report
        if not only_report_slug or only_report_slug == view.get_report_slug():
            view_perm = "%s.%s_view" % (
                view.get_base_model_name(),
                view.get_report_slug(),
            )
            if view_perm in user_reports or user.is_superuser:
                report_list.append(view)

    return report_list



from erp_framework.reporting.registry import register_report_view
from erp_framework.reporting.views import ReportView

from django.utils.translation import gettext_lazy as _

from .models import Expense, ExpenseTransaction


@register_report_view
class ExpenseTotalReport(ReportView):
    report_title = _("Expense Totals")

    # that's the report url endpoint
    # So this report ajax request will be sent reports/<namespace>/<report_slug>
    # this should be unique to the namespace
    report_slug = "balances"

    # here we define basic information for the report
    base_model = Expense

    # Where is the data to compute
    report_model = ExpenseTransaction

    group_by = "expense"
    columns = ["slug", "name", "__total__"]
    chart_settings = [
        {
            "type": "bar",
            "data_source": ["__total__"],
            "title_source": ["name"],
        },
        {
            "type": "pie",
            "data_source": ["__total__"],
            "title_source": ["name"],
        },
    ]


@register_report_view
class ExpenseDetailedStatement(ReportView):
    base_model = Expense
    report_model = ExpenseTransaction

    report_title = _("Expense detailed statement")
    columns = ["slug", "date", "type", "treasury__title", "expense__title", "value"]
    # if app_settings.RA_SUPPORT_LIABILITY:
    #     columns += ['liability__slug', 'liability__title']


@register_report_view
class ExpenseMovementTimeComparison(ReportView):
    report_title = _("Expenses Monthly Movements")
    base_model = Expense
    report_model = ExpenseTransaction
    group_by = "expense"
    time_series_pattern = "monthly"
    columns = ["slug", "name"]
    time_series_columns = ["__total__"]

    form_settings = {
        "default": True,
        "group_by": "treasury",
        "aggregate_on": "expense",
        "group_page_title": _("Expenses Monthly Movement"),
        "details_page_title": _("Expense Monthly Movement"),
        "time_series_pattern": "monthly",
        "time_series_scope": "both",
        "group_time_series_display": ["__total__"],
        "time_series_display": ["__total__"],
        "details_columns": ["treasury__slug", "treasury__title", "__total__"],
        "details_column_order": [
            "treasury__slug",
            "treasury__title",
            "__time_series__",
            "__total__",
        ],
        "group_columns": ["slug", "name", "__total__"],
        "group_column_order": [
            "treasury__slug",
            "treasury__title",
            "__time_series__",
            "__total__",
        ],
        "group_column_names": {
            "__total__": _("total expense movement"),
        },
        "time_series_TS_name": _("in"),
        "group_time_series_column_names": {
            "__total__": _("movement"),
        },
    }
    chart_settings = [
        {
            "id": "total_movement_bar",
            "type": "column",
            "title_source": "name",
            "data_source": ["__total__"],
            "series_names": [_("total movement")],
            "stacking": "normal",
        },
        # {
        #     'id': 'total_pie',
        #     'name': _('pie - sum'),
        #     'settings': {
        #         'chart_type': 'pie',
        #         'name': _('total movement comparison by month {expense}'),
        #         'sub_title': '{time_series_pattern} {date_verbose}',
        #         'y_sources': ['__total__'],
        #         'series_names': [_('total movement')],
        #         'plot_total': True
        #     }
        # },
        # {
        #     'id': 'balance_bar',
        #     'name': _('bar - detailed'),
        #     'settings': {
        #         'name': _('treasury expense comparison by month {expense}'),
        #         'sub_title': '{time_series_pattern} {date_verbose}',
        #         'chart_type': 'column',
        #         'y_sources': ['__total__'],
        #         'series_names': ['balance']
        #     }
        # },
    ]


@register_report_view
class ExpenseMovementMatrixComparison(ReportView):
    report_title = _("Expenses Comparison")
    base_model = Expense
    report_model = ExpenseTransaction

    group_by = "treasury"
    crosstab_model = "expense"
    crosstab_columns = ["__total__"]
    columns = ["name"]

    chart_settings = [
        {
            "type": "column",
            "data_source": ["__total__"],
            "plot_total": False,
            "title_source": "name",
        },
        {
            "type": "column",
            "data_source": ["__total__"],
            "plot_total": False,
            "title_source": "name",
            "stacking": "normal",
        },
        {
            "type": "pie",
            "data_source": ["__total__"],
            "plot_total": False,
            "title_source": "name",
        },
    ]

    #     'group_page_title': _('Expenses Comparison'),
    #     'details_page_title': _('Expenses Comparison'),
    #
    #     # 'time_series_pattern': 'monthly',
    #     # 'time_series_scope': 'both',
    #     # 'group_time_series_display': ['__total__'],
    #     # 'time_series_display': ['__total__'],
    #     'matrix': 'expense',
    #     'matrix_columns': ['__total__'],
    #     'matrix_columns_names': {
    #         '__total__': _('movement')
    #     },
    #
    #     'details_columns': ['treasury__slug', 'treasury__title', '__total__'],
    #     'details_column_order': ['treasury__slug', 'treasury__title', '__time_series__', '__total__'],
    #
    #     'group_columns': ['slug', 'name', '__total__'],
    #     'group_column_order': ['treasury__slug', 'treasury__title', '__time_series__', '__total__'],
    #     'group_column_names': {
    #         '__total__': _('total expense movement'),
    #     },
    #
    #     'time_series_TS_name': _('in'),
    #
    #     'group_time_series_column_names': {
    #         '__total__': _('movement'),
    #     },
    #
    # }



# Generated by Django 2.2.12 on 2020-06-01 20:49

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import erp_framework.base.models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("contenttypes", "0002_remove_content_type_name"),
        ("treasury", "__first__"),
    ]

    operations = [
        migrations.CreateModel(
            name="Expense",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "creation_date",
                    models.DateTimeField(
                        default=django.utils.timezone.now,
                        verbose_name="creation date and time",
                    ),
                ),
                (
                    "lastmod",
                    models.DateTimeField(
                        db_index=True, verbose_name="last modification"
                    ),
                ),
                (
                    "slug",
                    models.SlugField(
                        blank=True,
                        help_text="For fast recall",
                        unique=True,
                        verbose_name="refer code",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        db_index=True, max_length=255, unique=True, verbose_name="name"
                    ),
                ),
                (
                    "notes",
                    models.TextField(blank=True, null=True, verbose_name="notes"),
                ),
                (
                    "lastmod_user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="expense_expense_lastmod_related",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="last modification by",
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="expense_expense_related",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="owner",
                    ),
                ),
                (
                    "polymorphic_ctype",
                    models.ForeignKey(
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="polymorphic_expense.expense_set+",
                        to="contenttypes.ContentType",
                    ),
                ),
            ],
            options={
                "verbose_name": "Expense",
                "verbose_name_plural": "Expenses",
            },
            bases=(erp_framework.base.models.DiffingMixin, models.Model),
        ),
        migrations.CreateModel(
            name="ExpenseTransaction",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("slug", models.SlugField(blank=True, verbose_name="refer code")),
                ("date", models.DateTimeField(db_index=True, verbose_name="date")),
                ("type", models.CharField(db_index=True, max_length=30)),
                (
                    "notes",
                    models.TextField(blank=True, null=True, verbose_name="notes"),
                ),
                (
                    "value",
                    models.DecimalField(
                        decimal_places=2, default=0, max_digits=19, verbose_name="value"
                    ),
                ),
                (
                    "creation_date",
                    models.DateTimeField(
                        default=django.utils.timezone.now,
                        verbose_name="creation date and time",
                    ),
                ),
                (
                    "lastmod",
                    models.DateTimeField(
                        db_index=True, verbose_name="last modification"
                    ),
                ),
                (
                    "expense",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="expense.Expense",
                        verbose_name="Expense",
                    ),
                ),
                (
                    "lastmod_user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="expense_expensetransaction_lastmod_related",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="last modification by",
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="expense_expensetransaction_related",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="owner",
                    ),
                ),
                (
                    "polymorphic_ctype",
                    models.ForeignKey(
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="polymorphic_expense.expensetransaction_set+",
                        to="contenttypes.ContentType",
                    ),
                ),
                (
                    "treasury",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="treasury.Treasury",
                        verbose_name="Treasury",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(erp_framework.base.models.DiffingMixin, models.Model),
        ),
    ]



from __future__ import unicode_literals

from collections import OrderedDict

from django.contrib.admin.sites import AlreadyRegistered, NotRegistered
from django.core.exceptions import ImproperlyConfigured


class ReportRegistry(object):
    def __init__(self):
        super(ReportRegistry, self).__init__()
        self._registry = OrderedDict()
        self._slugs_registry = []
        self._store = {}
        self._base_models = []

    def register(self, report_class):
        """
        Register report class
        :param report_class:
        :return:
        """
        if not report_class.hidden:

            try:
                namespace = report_class.base_model._meta.model_name
            except AttributeError:
                raise ImproperlyConfigured("Can not access base_model, is it set on class %s?" % report_class)
            try:
                if not report_class.report_title:
                    raise AttributeError
            except AttributeError:
                raise ImproperlyConfigured('Report %s is missing a `report_title`' % report_class)
            # try:
            #     assert type(report_class.form_settings) is dict
            # except (AttributeError, AssertionError):
            #     raise ImproperlyConfigured(
            #         'Report %s is missing a `form_settings` or form_settings is not a dict' % report_class)
            if not report_class.get_report_model():
                raise ImproperlyConfigured(
                    'Report %s is missing a `report_model`' % report_class)
            # if report_class.must_exist_filter and not report_class.header_report:
            #     raise ImproperlyConfigured('%s: Must specify a view class or function in `header_report` '
            #                                'if `must_exist_filter` is set' % report_class)

            self._register_report(report_class, namespace)

    def _register_report(self, report, namespace):
        """
        Actual registry function, recursive
        :param report:
        :param namespace:
        :return:
        """
        namespace_existing = namespace in self._registry
        full_name = '%s_%s' % (namespace, report.get_report_slug())
        if namespace_existing:
            if report in self._registry[namespace]:
                raise AlreadyRegistered('This report class is already registered %s' % report)

            if full_name in self._slugs_registry:
                raise AlreadyRegistered('report slug `%s` is already registered for `%s`' %
                                        (report.get_report_slug(), namespace))

            self._registry[namespace].append(report)
        else:
            self._registry[namespace] = [report]
            if report.base_model not in self._base_models:
                self._base_models.append(report.base_model)
        self._slugs_registry.append(full_name)
        self._store[full_name] = report

    def unregister(self, report_class, w_other_namespaces=True):
        self._unregister(report_class, report_class.namespace)
        if w_other_namespaces:
            other_namespaces = report_class.other_namespaces or []
            for e in other_namespaces:
                self._unregister(report_class, e)

    def unregister_namespace(self, namespace):
        reports = self._registry.pop(namespace, [])
        for r in reports:
            slug_id = '%s_%s' % (namespace, r.get_reprot_slug())
            self._store.pop(slug_id)
            self._slugs_registry.remove(slug_id)
        return reports

    def _unregister(self, report_class, namespace):
        if type(self._registry[namespace]) is list:
            if report_class not in self._registry[namespace]:
                raise NotRegistered('This report is not registered')
            self._registry[namespace].remove(report_class)
            slug_id = '%s_%s' % (report_class.namespace, report_class.get_report_slug())
            self._slugs_registry.remove(slug_id)
            self._base_models.remove(report_class.base_model)

    def get_report_classes_by_namespace(self, namespace):
        if namespace in self._registry:
            return self._registry[namespace]
        return []
        # else:
        #     raise NotRegistered(namespace)

    def get_all_reports(self):
        reports = []
        for namespace in self._registry:
            reports += list(self._registry[namespace])
        return reports

    def get(self, namespace, report_slug):
        slug_id = '%s_%s' % (namespace, report_slug)
        try:
            return self._store[slug_id.lower()]
        except KeyError:
            raise NotRegistered(
                "Report '%s' base model '%s' not found, Did you register it? If yes, then maybe it's has different base model ?" % (
                    report_slug, namespace))

    def get_base_models(self):
        return self._base_models

    def get_base_models_with_reports(self):
        output = []
        for i, k in enumerate(self._registry.keys()):
            v = self._registry[k]
            output.append((self._base_models[i], v))
        return output


report_registry = ReportRegistry()


def register_report_view(report_class=None, condition=None):
    if report_class:
        report_registry.register(report_class)
        return report_class

    def wrapper(report_class):
        if callable(condition):
            if not condition():
                return report_class
        report_registry.register(report_class)
        return report_class

    return wrapper



from django.apps import AppConfig

from django.utils.translation import gettext_lazy as _


class ERPFrameworkConfig(AppConfig):
    name = "erp_framework"
    verbose_name = _("ERP framework")

    # def ready(self):
    from django.db.models.signals import post_migrate

    #     super(RaConfig, self).ready()
    #     from .utils.permissions import create_report_permissions
    #     from . import checks
    #
    #     post_migrate.connect(
    #         create_report_permissions,
    #         sender=self,
    #         dispatch_uid="erp_framework.utils.permissions.create_reports_permissions",
    #     )



from erp_framework.admin.admin import EntityAdmin, TransactionAdmin, erp_admin_site
from .models import Expense, ExpenseTransaction


class ExpenseAdmin(EntityAdmin):
    pass


class ExpenseTransactionAdmin(TransactionAdmin):
    fields = ("slug", "date", "treasury", "expense", "value", "notes")


erp_admin_site.register(Expense, ExpenseAdmin)
erp_admin_site.register(ExpenseTransaction, ExpenseTransactionAdmin)



from erp_framework.base.models import EntityModel
from django.utils.translation import gettext_lazy as _


class Treasury(EntityModel):
    class Meta:
        verbose_name = _("Treasury")
        verbose_name_plural = _("Treasuries")



# Generated by Django 2.2.12 on 2020-07-11 12:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('expense', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='expense',
            name='notes',
            field=models.TextField(blank=True, null=True, verbose_name='Notes'),
        ),
        migrations.AlterField(
            model_name='expense',
            name='slug',
            field=models.SlugField(blank=True, help_text='For fast recall', unique=True, verbose_name='Identifier slug'),
        ),
        migrations.AlterField(
            model_name='expense',
            name='name',
            field=models.CharField(db_index=True, max_length=255, unique=True, verbose_name='Name'),
        ),
    ]



from django.conf import settings
from django.core.checks import Error, register, Warning
from django.utils.module_loading import import_string


@register()
def check_ra_settings(app_configs, **kwargs):
    errors = []
    if (
        "erp_framework.base.context_processors.global_info"
        not in settings.TEMPLATES[0]["OPTIONS"]["context_processors"]
    ):
        errors.append(
            Error(
                "context erp_framework.base.context_processors.global_info is missing",
                hint="add erp_framework.base.context_processors.global_info to context_processor",
                obj="settings",
                id="erp_framework.E002",
            )
        )
    return errors


@register()
def check_crequest(app_configs=None, **kwargs):
    errors = []
    if "crequest" not in settings.INSTALLED_APPS:
        errors.append(
            Error(
                "crequest app is missing",
                hint="Add `crequest` to INSTALLED_APPS",
                obj="settings",
                id="erp_framework.E003",
            )
        )
    if "crequest.middleware.CrequestMiddleware" not in settings.MIDDLEWARE:
        errors.append(
            Error(
                "crequest middleware is missing",
                hint='Add "crequest.middleware.CrequestMiddleware" to MIDDLEWARE',
                obj="settings",
                id="erp_framework.E003",
            )
        )
    return errors


@register()
def check_initial_settings(app_configs, **kwargs):
    errors = []
    if (
        "django.template.context_processors.i18n"
        not in settings.TEMPLATES[0]["OPTIONS"]["context_processors"]
    ):
        errors.append(
            Error(
                "django.template.context_processors.i18n is missing",
                hint='Add "django.template.context_processors.i18n" to context_processors',
                obj="settings",
                id="erp_framework.E003",
            )
        )
    if (
        "django.template.context_processors.static"
        not in settings.TEMPLATES[0]["OPTIONS"]["context_processors"]
    ):
        errors.append(
            Error(
                "django.template.context_processors.static is missing",
                hint='Add "django.template.context_processors.static" to context_processors',
                obj="settings",
                id="erp_framework.E003",
            )
        )

    return errors



from __future__ import absolute_import

from django import forms
from django.contrib.admin.widgets import RelatedFieldWidgetWrapper
from django.forms import SplitDateTimeWidget
from django.utils.safestring import mark_safe


class AdminSplitDateTimeNoBr(forms.SplitDateTimeWidget):
    pass


class RaRelatedFieldWidgetWrapper(RelatedFieldWidgetWrapper):
    pass


# class RaBootstrapDateTime(SplitDateTimeWidget):
#     def render(self, name, value, attrs=None, renderer=None):
#         if self.is_localized:
#             for widget in self.widgets:
#                 widget.is_localized = self.is_localized
#         # value is a list of values, each corresponding to a widget
#         # in self.widgets.
#         if not isinstance(value, list):
#             value = self.decompress(value)
#         output = []
#         final_attrs = self.build_attrs(attrs)
#         id_ = final_attrs.get('id', None)
#         for i, widget in enumerate(self.widgets):
#             try:
#                 widget_value = value[i]
#             except IndexError:
#                 widget_value = None
#             if id_:
#                 final_attrs = dict(final_attrs, id='%s_%s' % (id_, i))
#             output.append(widget.render(name + '_%s' % i, widget_value, final_attrs))
#         return_html = self.get_foundation_structure() % {'datewidget': output[0], 'timewidget': output[1]}
#         return mark_safe(return_html)
#
#     def get_foundation_structure(self):
#         html = '<div class="row"> \
#           <div class="col-sm-6"> \
#           %(datewidget)s \
#           </div>\
#           <div class="col-sm-6"> \
#             %(timewidget)s \
#           </div>\
#         </div>'
#
#         return html



from __future__ import unicode_literals

from django import template
from django.template.defaultfilters import capfirst
from django.template.loader import get_template, render_to_string
from django.utils.module_loading import import_string
from django.utils.safestring import mark_safe

from erp_framework.base import app_settings

register = template.Library()


@register.simple_tag(takes_context=True)
def render_navigation_menu(context):
    navigation_class = import_string(app_settings.RA_NAVIGATION_CLASS)
    request = context["request"]
    admin_site = context["admin_site"]
    return mark_safe(navigation_class.get_menu(context, request, admin_site))


@register.simple_tag(takes_context=True)
def render_reports_menu(context):
    request = context["request"]
    is_in_reports = False
    active_base_model = ""
    if request.path.startswith("/reports/"):
        is_in_reports = True
        active_base_model = [x for x in request.path.split("/") if x][1]

    from erp_framework.reporting.registry import report_registry

    base_models = report_registry.get_base_models_with_reports()
    if base_models:
        t = get_template(f"erp_framework/reports_menu.html")
        output = render_to_string(
            "erp_framework/reports_menu.html",
            {
                "base_models_reports_tuple": base_models,
                "is_report": context.get("is_report", False),
                "base_model": context.get("base_model", False),
                "report_slug": context.get("report_slug", False),
                "current_base_model_name": context.get(
                    "current_base_model_name", False
                ),
            },
            request,
        )
        return mark_safe(output)
    return ""


@register.simple_tag(takes_context=True)
def get_report(context, base_model, report_slug):
    from erp_framework.reporting.registry import report_registry

    return report_registry.get(namespace=base_model, report_slug=report_slug)


@register.simple_tag()
def get_model_verbose_name_plural(model):
    return capfirst(model._meta.verbose_name_plural)


@register.simple_tag(takes_context=True)
def get_report_active_class(context, base_model, css_class=None):
    css_class = css_class or "active"
    current_base_model_name = context["current_base_model_name"]
    return (
        css_class
        if current_base_model_name == base_model._meta.model_name.lower()
        else ""
    )



from django import apps
from django.utils.translation import gettext_lazy as _


class ExpenseConfig(apps.AppConfig):
    name = 'erp_framework.erp.expense'
    verbose_name = _("Expense")

    def ready(self):
        super(ExpenseConfig, self).ready()
        from . import reports




from __future__ import unicode_literals

import logging
from functools import update_wrapper

from django.apps import apps
from django.contrib.admin import AdminSite
from django.http import JsonResponse
from django.shortcuts import render
from django.template.response import TemplateResponse
from django.urls import re_path as url, include
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _

from erp_framework.admin.helpers import get_each_context
from erp_framework.base import app_settings
from erp_framework.base.helpers import get_from_list

# from erp_framework.top_search.views import TopSearchView

logger = logging.getLogger(__name__)
CACHE_DURATION = 0

MSG_ADDED_SUCCESSFULLY = _('The %(name)s "%(obj)s" was added successfully.')
MSG_CHANGED_SUCCESSFULLY = _('The %(name)s "%(obj)s" was changed successfully.')
MSG_YOU_MAY_ADD = _("You may add another %(name)s below")
MSG_YOU_MAY_CHANGE = _("You may change it again below.")


def get_report_list_class(request, base_model):
    from erp_framework.base.app_settings import RA_REPORT_LIST_MAP
    from erp_framework.reporting.views import ReportList

    klass = RA_REPORT_LIST_MAP.get(base_model, False)
    if klass:
        klass = import_string(klass)
        if callable(klass):
            return klass(request, base_model=base_model)
    else:
        klass = ReportList

    return klass.as_view()(request, base_model=base_model)


def get_report_view(request, base_model, report_slug):
    from erp_framework.reporting.registry import report_registry

    klass = report_registry.get(base_model, report_slug)
    return klass.as_view()(request)


class RaAdminSiteBase(AdminSite):
    site_title = app_settings.RA_ADMIN_SITE_TITLE
    site_header = app_settings.RA_ADMIN_SITE_HEADER
    index_title = app_settings.RA_ADMIN_INDEX_TITLE

    index_template = app_settings.RA_ADMIN_INDEX_TEMPLATE
    app_index_template = app_settings.RA_ADMIN_APP_INDEX_TEMPLATE
    login_template = app_settings.RA_ADMIN_LOGIN_TEMPLATE

    # logout_template = app_settings.RA_ADMIN_LOGGED_OUT_TEMPLATE

    def get_urls(self):
        # from erp_framework.utils.views import access_denied
        def wrap(view, cacheable=False):
            def wrapper(*args, **kwargs):
                return self.admin_view(view, cacheable)(*args, **kwargs)

            return update_wrapper(wrapper, view)

        urls = super(RaAdminSiteBase, self).get_urls()
        help_center = [
            url(r"^i18n/", include("django.conf.urls.i18n")),
        ]

        settings_update = [
            url(r"^manifest/$", self.manifest_view, name="manifest"),
            url(r"^sw.js$", self.service_worker_view, name="service-worker"),
        ]

        urlpatterns = [
            url(
                r"^reports/(?P<base_model>[\w-]+)/$",
                get_report_list_class,
                name="report_list",
            ),
            url(
                r"^reports/(?P<base_model>[\w-]+)/(?P<report_slug>[\w-]+)/$",
                get_report_view,
                name="report",
            ),
            # new from sites
            # path('top-search/', TopSearchView.as_view(), name='top-search'),
            # path('access-denied/', access_denied, name='access-denied'),
        ]

        return help_center + settings_update + urlpatterns + urls

    def service_worker_view(self, request):
        return render(
            request,
            f"erp_framework/service-worker.js.html",
            content_type="application/javascript",
        )

    def manifest_view(self, request):
        json = {
            "short_name": "ERP Framework System",
            "name": "ERP Framework System",
            "icons": [
                {
                    "src": "/static/ra/images/ra_systems.png",
                    "type": "image/png",
                    "sizes": "147x147",
                },
                # {
                #     "src": "launcher-icon-2x.png",
                #     "type": "image/png",
                #     "sizes": "96x96"
                # },
                # {
                #     "src": "launcher-icon-4x.png",
                #     "type": "image/png",
                #     "sizes": "192x192"
                # }
            ],
            "start_url": "../?launcher=true",
            "display": "standalone",
            "theme_color": "#000066",
            "background_color": "#fff",
        }
        return JsonResponse(json, status=200)

    def __init__(self, name="admin"):
        super(RaAdminSiteBase, self).__init__(name)

        self._registry_names = {}  # holds model name -> ModelAdmin instances
        #                  l-> Model
        self._registered_models_CT = []

    def _fill_registry_names(self):
        """
        Fills the `self._registry_names`
        """
        for model in self._registry.keys():
            try:
                model_name = model.get_class_name().lower()
            except AttributeError:
                model_name = model._meta.model_name.lower()
            self._registry_names[model_name] = {
                "admin": self._registry[model],
                "model": model,
            }

    def app_index(self, request, app_label, extra_context=None):
        app_name = apps.get_app_config(app_label).verbose_name
        context = self.each_context(request)
        app_list = context["app_list"]
        current_app_list = get_from_list(False, app_list, "app_label", app_label)
        context.update(
            dict(
                title=_("%(app)s administration") % {"app": app_name},
                # current_app_list=[app_dict],
                current_app_list=[current_app_list],
                app_label=app_label,
                app_name=app_name,
            )
        )
        context.update(extra_context or {})
        request.current_app = self.name

        return TemplateResponse(
            request,
            self.app_index_template
            or ["admin/%s/app_index.html" % app_label, "admin/app_index.html"],
            context,
        )

    def index(self, request, extra_context=None):
        extra_context = extra_context or {}
        extra_context["opts"] = {"app_name": "home"}
        extra_context["is_index"] = True
        extra_context["sidebar_status"] = "expanded"
        context = dict(
            self.each_context(request),
            name=self.index_title,
        )
        context.update(extra_context or {})

        request.current_app = self.name

        return TemplateResponse(
            request, self.index_template or "admin/index.html", context
        )

    def each_context(self, request):
        context = super(RaAdminSiteBase, self).each_context(request)
        context["RA_ADMIN_SITE_NAME"] = app_settings.RA_ADMIN_SITE_NAME
        context.update(get_each_context(request, self))
        return context

    def get_admin_by_model_name(self, model_name):
        """
        Get the model admin of a model by its name
        :param model_name:
        :return: ModalAdmin Instance or None
        """
        if not self._registry_names:
            self._fill_registry_names()
        return self._registry_names.get(model_name, None)

    def login(self, request, extra_context=None):
        extra_context = extra_context or {}
        # extra_context["SHOW_LANGUAGE_SELECTOR"] = True
        return super(RaAdminSiteBase, self).login(request, extra_context)



from django import apps
from django.db.models.signals import post_migrate


class ReportAppConfig(apps.AppConfig):
    name = "erp_framework.reporting"

    def ready(self):
        from erp_framework.utils.permissions import create_report_permissions

        super().ready()
        post_migrate.connect(create_report_permissions, sender=self)



from __future__ import unicode_literals

import logging
from functools import update_wrapper

from crequest.middleware import CrequestMiddleware
from django.contrib import admin
from django.contrib.admin import helpers
from django.contrib.admin.exceptions import DisallowedModelAdminToField
from django.contrib.admin.options import (
    IS_POPUP_VAR,
    TO_FIELD_VAR,
    get_content_type_for_model,
)
from django.contrib.admin.templatetags.admin_urls import add_preserved_filters
from django.contrib.admin.utils import quote, unquote, flatten_fieldsets
from django.contrib.admin.widgets import AdminTextareaWidget
from django.contrib.auth import get_permission_codename
from django.contrib.auth.admin import UserAdmin, GroupAdmin
from django.contrib.auth.models import User, Group
from django.contrib.contenttypes.admin import GenericTabularInline
from django.core.exceptions import PermissionDenied, ImproperlyConfigured
from django.db import models
from django.db.models import DecimalField, ForeignKey, TextField
from django.db.models.base import ModelBase
from django.dispatch import Signal
from django.forms import all_valid
from django.forms.widgets import TextInput, NumberInput
from django.http import HttpResponseRedirect, Http404
from django.template.defaultfilters import capfirst
from django.template.response import TemplateResponse
from django.urls import reverse, path
from django.utils.decorators import method_decorator
from django.utils.html import escape
from django.utils.module_loading import import_string
from django.utils.safestring import mark_safe
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _, gettext
from django.views.decorators.csrf import csrf_protect
from reversion.admin import VersionAdmin
from tabular_permissions.admin import (
    UserTabularPermissionsMixin,
    GroupTabularPermissionsMixin,
)

from erp_framework.base.helpers import flatten_list
from erp_framework.admin.forms import RaUserChangeForm
from erp_framework.reporting.printing import HTMLPrintingClass
from erp_framework.utils.views import (
    get_typed_reports_for_templates,
    get_typed_reports_map,
    apply_order_to_typed_reports,
)
from .base import RaAdminSiteBase
from ..base import app_settings
from ..base.widgets import RaRelatedFieldWidgetWrapper

csrf_protect_m = method_decorator(csrf_protect)

default_fields = (("name", "slug"), "notes")

default_exclude = ("owner", "creation_date", "lastmod", "lastmod_user")

default_list_display = ("name", "slug", "notes")

logger = logging.getLogger(__name__)

from django.contrib.admin.views.main import ChangeList

# todo bring back
# changeform_saved = Signal(providing_args=["instance", "created", 'using'])
changeform_saved = Signal()


def get_reports_map(model_name, user, request, order_list=None):
    """
    Gets ReportViews associated with a specific model ie. namespace
    :param model_name: or namespace defined in ReportView namespace
    :param user: User to check for permissions
    :param order_list: list to order the reports in comparision to

    :return: a dictionary with two values `slugs` and `reports`
        slugs: map to a list of report slugs
        reports: map to a list of ReportView classes
    """

    # todo refine with erp_framework.views.ReportList View
    order_list = list(order_list or [])
    reports = get_typed_reports_for_templates(model_name, user, request)
    reports = apply_order_to_typed_reports(reports, order_list)
    return get_typed_reports_map(typed_reports=reports)


class RaChangeList(ChangeList):
    def __init__(
        self,
        request,
        model,
        list_display,
        list_display_links,
        list_filter,
        date_hierarchy,
        search_fields,
        list_select_related,
        list_per_page,
        list_max_show_all,
        list_editable,
        model_admin,
        sortable_by,
    ):
        super(RaChangeList, self).__init__(
            request,
            model,
            list_display,
            list_display_links,
            list_filter,
            date_hierarchy,
            search_fields,
            list_select_related,
            list_per_page,
            list_max_show_all,
            list_editable,
            model_admin,
            sortable_by,
        )
        self.request = request  # Add request to the class
        if self.is_popup:
            name = gettext("Select %s")
        else:
            name = gettext("%s")
        self.name = name % str(self.opts.verbose_name_plural)

        self.no_records_message = model_admin.no_records_message

    # def url_for_result(self, result):
    #     """
    #     Override parent to handle correct url mapping
    #     :param result:
    #     :return:
    #     """
    #     if self.model_admin.has_view_permission(self.request):
    #         pk = getattr(result, self.pk_attname)
    #         return reverse('%s:%s_%s_view' % (app_settings.RA_ADMIN_SITE_NAME, self.opts.app_label,
    #                                           self.opts.model_name),
    #                        args=(quote(pk),),
    #                        current_app=self.model_admin.admin_site.name)
    #     elif self.model_admin.has_change_permission(self.request):
    #         pk = getattr(result, self.pk_attname)
    #         return reverse('%s:%s_%s_change' % (app_settings.RA_ADMIN_SITE_NAME, self.opts.app_label,
    #                                             self.opts.model_name),
    #                        args=(quote(pk),),
    #                        current_app=self.model_admin.admin_site.name)
    #     elif self.model_admin.has_add_permission(self.request):
    #         return reverse('%s:%s_%s_add' % (app_settings.RA_ADMIN_SITE_NAME, self.opts.app_label,
    #                                          self.opts.model_name),
    #                        current_app=self.model_admin.admin_site.name)
    #     return False


class RaAdminSite(RaAdminSiteBase):
    pass


site_class = import_string(app_settings.RA_ADMIN_SITE_CLASS)
erp_admin_site = site_class(name=app_settings.RA_ADMIN_SITE_NAME)


class RaThemeMixin:
    change_form_template = f"{app_settings.RA_THEME}/change_form.html"
    change_list_template = f"{app_settings.RA_THEME}/change_list.html"
    delete_confirmation_template = f"{app_settings.RA_THEME}/delete_confirmation.html"
    delete_selected_confirmation_template = (
        f"{app_settings.RA_THEME}/delete_selected_confirmation.html"
    )
    add_form_template = f"{app_settings.RA_THEME}/change_form.html"

    recover_form_template = f"erp_framework/reversion/recover_form.html"
    revision_form_template = f"erp_framework/reversion/revision_form.html"
    object_history_template = f"erp_framework/reversion/object_history.html"
    recover_list_template = f"erp_framework/reversion/recover_list.html"

    view_template = None  # Defaults to f'{app_settings.RA_THEME}/view.html'


class AdminViewMixin(admin.ModelAdmin):
    enable_view_view = True
    view_fields = ()
    view_template = "erp_framework/view.html"

    def get_stats_icon(self, obj):
        url = reverse(
            "%s:%s_%s_view"
            % ("admin", self.model._meta.app_label, self.model._meta.model_name),
            args=(quote(obj.pk),),
            current_app=self.admin_site.name,
        )

        # todo change the str obj
        view_link = (
            '<a href="%s" data-popup="tooltip" name="%s %s" data-placement="bottom"> '
            '<i class="fas fa-chart-line"></i> </a>'
            % (url, capfirst(_("Statistics for")), str(obj))
        )
        return mark_safe(view_link)

    get_stats_icon.short_description = _("Stats")

    def get_view_fields(self, request, obj=None):
        return (
            self.view_fields
            or [x for x in flatten_list(self.fields)]
            or self.get_fields(request, obj)
        )

    def get_view_title(self, request, obj=None):
        return _("View %s") % str(obj)
        # return self.get_title(request, obj)

    def get_urls(self):
        urls = super().get_urls()
        info = self.model._meta.app_label, self.model._meta.model_name

        my_urls = [
            path(
                "<path:object_id>/view/",
                self.admin_site.admin_view(self.view_view),
                name="%s_%s_view" % info,
            ),
        ]
        return my_urls + urls

    def has_view_permission(self, request, obj=None):
        if not self.enable_view_view:
            return False
        opts = self.opts
        codename = get_permission_codename("view", opts)
        return request.user.has_perm("%s.%s" % (opts.app_label, codename))

    def view_view(self, request, object_id, form_url="", extra_context=None):
        extra_context = extra_context or {}
        extra_context["has_add_permission"] = self.has_add_permission(request)

        to_field = request.POST.get(TO_FIELD_VAR, request.GET.get(TO_FIELD_VAR))
        if to_field and not self.to_field_allowed(request, to_field):
            raise DisallowedModelAdminToField(
                "The field %s cannot be referenced." % to_field
            )

        model = self.model
        opts = model._meta

        obj = self.get_object(request, unquote(object_id), to_field)

        if not self.has_view_permission(request, obj):
            raise PermissionDenied

        if obj is None:
            return self._get_obj_does_not_exist_redirect(request, opts, object_id)

        data = []
        view_fields = self.get_view_fields(request, obj)
        for field_name in view_fields:
            # field = self.get_field(request, obj, field_name)
            field = model._meta.get_field(field_name)
            if field:
                data.append((field.verbose_name, field.value_from_object(obj)))

        context = dict(
            self.admin_site.each_context(request),
            name=obj,
            app_label=opts.app_label,
            object_id=object_id,
            original=obj,
            is_popup=(IS_POPUP_VAR in request.POST or IS_POPUP_VAR in request.GET),
            to_field=to_field,
            # media=media,
            title=self.get_view_title(request, obj),
            data=data,
            preserved_filters=self.get_preserved_filters(request),
        )

        context.update(extra_context or {})

        opts = self.model._meta
        app_label = opts.app_label
        preserved_filters = self.get_preserved_filters(request)
        form_url = add_preserved_filters(
            {"preserved_filters": preserved_filters, "opts": opts}, form_url
        )
        view_on_site_url = self.get_view_on_site_url(obj)
        context.update(
            {
                # 'add': add,
                # 'change': not add,
                "has_add_permission": self.has_add_permission(request),
                "has_change_permission": self.has_change_permission(request, obj),
                "has_view_permission": self.has_view_permission(request, obj),
                "has_delete_permission": self.has_delete_permission(request, obj),
                "has_file_field": True,
                "has_absolute_url": view_on_site_url is not None,
                "absolute_url": view_on_site_url,
                "form_url": form_url,
                "opts": opts,
                "content_type_id": get_content_type_for_model(self.model).pk,
                "save_as": self.save_as,
                "save_on_top": self.save_on_top,
                "to_field_var": TO_FIELD_VAR,
                "is_popup_var": IS_POPUP_VAR,
                "app_label": app_label,
            }
        )

        request.current_app = self.admin_site.name
        return TemplateResponse(
            request,
            self.view_template
            or [
                "erp_framework/%s/%s/view.html" % (opts.app_label, opts.model_name),
                "erp_framework/%s/view.html" % opts.app_label,
                "erp_framework/view.html",
                f"{app_settings.RA_THEME}/view.html",
            ],
            context,
        )


class EntityAdmin(RaThemeMixin, AdminViewMixin, VersionAdmin):
    # ModelAdmin Attributes
    view_on_site = False
    list_per_page = 25
    save_on_top = True
    list_select_related = True
    fields = default_fields
    list_display = ("get_enhanced_obj_title", "slug", "notes", "get_history_link")
    list_display_links = ("slug",)
    date_hierarchy = "creation_date"

    # Custom templates

    history_latest_first = True
    search_fields = ["name", "slug"]
    formfield_overrides = {
        # DateTimeField: {'widget': AdminSplitDateTime},
        # DecimalField: {'widget': NumberInput},
        # ForeignKey: {'widget': RaBootstrapForeignKeyWidget},
        TextField: {"widget": AdminTextareaWidget(attrs={"rows": 2})}
    }

    # Ra New Attributes
    # ------------------
    description = None  # Description of the model to be displayed in hte changelist filter under the name
    enable_next_serial = True  # Enable the default slug to be a serial number
    enable_view_view = True  # If False there is no "DetailView ie Stats view"
    print_template = None
    no_records_message = _("No Data")

    typed_reports_order_list = None
    # new attrs:
    print_settings = {}
    print_title = ""
    print_class = HTMLPrintingClass

    def reversion_register(self, model, **kwargs):
        kwargs["for_concrete_model"] = False
        super(EntityAdmin, self).reversion_register(model, **kwargs)

    def get_enhanced_obj_title(self, obj):
        request = CrequestMiddleware.get_request()
        links = []
        pk_attname = self.model._meta.pk.attname
        pk = getattr(obj, pk_attname)
        main_url = False
        if self.has_change_permission(request, obj):
            url = reverse(
                "%s:%s_%s_change"
                % (
                    app_settings.RA_ADMIN_SITE_NAME,
                    self.model._meta.app_label,
                    self.model._meta.model_name,
                ),
                args=(quote(pk),),
                current_app=self.admin_site.name,
            )
            if not main_url:
                main_url = url
            view_link = (
                '<a href="%s" data-popup="tooltip" name="%s" data-placement="bottom">'
                ' <i class="fas fa-edit"></i> </a>' % (url, capfirst(_("change")))
            )
            links.append(view_link)
            links = "<span class='go-to-change-form'>%s</span>" % "".join(links) + ""
            obj_link_title = "<a href='%s'>%s</a>" % (main_url, obj.name)
            return mark_safe("%s %s" % (obj_link_title, links))

        return obj.name

    def get_history_link(self, obj):
        info = self.model._meta.app_label, self.model._meta.model_name
        url = reverse("erp_admin::%s_%s_history" % info, args=(obj.pk,))

        return mark_safe(
            """<a href="%s" class="legitRipple" data-popup="tooltip" name="%s">
                        <i class="fas fa-history text-indigo-800"></i>
                        <span class="legitRipple-ripple"></span></a>"""
            % (url, _("History"))
        )

    get_history_link.short_description = _("History")

    get_enhanced_obj_title.short_description = _("name")
    get_enhanced_obj_title.admin_order_field = "name"

    # Permissions
    # def has_view_permission(self, request, obj=None):
    #     if not self.enable_view_view:
    #         return False
    #     opts = self.opts
    #     codename = get_permission_codename('view', opts)
    #     return request.user.has_perm("%s.%s" % (opts.app_label, codename))
    #
    # def view_view(self, request, object_id, form_url='', extra_context=None):
    #     extra_context = extra_context or {}
    #     extra_context['has_add_permission'] = self.has_add_permission(request)
    #     '''
    #         Code copied from ChangeForm
    #     '''
    #
    #     to_field = request.POST.get(TO_FIELD_VAR, request.GET.get(TO_FIELD_VAR))
    #     if to_field and not self.to_field_allowed(request, to_field):
    #         raise DisallowedModelAdminToField("The field %s cannot be referenced." % to_field)
    #
    #     model = self.model
    #     opts = model._meta
    #     add = object_id is None
    #
    #     if add:
    #         if not self.has_add_permission(request):
    #             raise PermissionDenied
    #         obj = None
    #
    #     else:
    #         obj = self.get_object(request, unquote(object_id), to_field)
    #
    #         if not self.has_view_permission(request, obj):
    #             raise PermissionDenied
    #
    #         if obj is None:
    #             raise Http404(_('%(name)s object with primary key %(key)r does not exist.') % {
    #                 'name': str(opts.verbose_name), 'key': escape(object_id)})
    #
    #             # if request.method == 'POST' and "_saveasnew" in request.POST:
    #             #     return self.add_view(request, form_url=reverse('admin:%s_%s_add' % (
    #             #         opts.app_label, opts.model_name), current_app=self.admin_site.name))
    #     context = dict(self.admin_site.each_context(request),
    #                    name=obj,
    #                    app_label=opts.app_label,
    #                    object_id=object_id,
    #                    original=obj,
    #                    is_popup=(IS_POPUP_VAR in request.POST or
    #                              IS_POPUP_VAR in request.GET),
    #                    to_field=to_field,
    #                    # media=media,
    #                    preserved_filters=self.get_preserved_filters(request),
    #                    )
    #
    #     context.update(extra_context or {})
    #
    #     opts = self.model._meta
    #     app_label = opts.app_label
    #     preserved_filters = self.get_preserved_filters(request)
    #     form_url = add_preserved_filters({'preserved_filters': preserved_filters, 'opts': opts}, form_url)
    #     view_on_site_url = self.get_view_on_site_url(obj)
    #     context.update({
    #         'add': add,
    #         'change': not add,
    #         'has_add_permission': self.has_add_permission(request),
    #         'has_change_permission': self.has_change_permission(request, obj),
    #         'has_delete_permission': self.has_delete_permission(request, obj),
    #         'has_file_field': True,
    #         'has_absolute_url': view_on_site_url is not None,
    #         'absolute_url': view_on_site_url,
    #         'form_url': form_url,
    #         'opts': opts,
    #         'content_type_id': get_content_type_for_model(self.model).pk,
    #         'save_as': self.save_as,
    #         'save_on_top': self.save_on_top,
    #         'to_field_var': TO_FIELD_VAR,
    #         'is_popup_var': IS_POPUP_VAR,
    #         'app_label': app_label,
    #     })
    #
    #     request.current_app = self.admin_site.name
    #     return TemplateResponse(request, self.view_template or [
    #         "ra/%s/%s/view.html" % (opts.app_label, opts.model_name),
    #         "ra/%s/view.html" % opts.app_label,
    #         'ra/view.html',
    #         f"{app_settings.RA_THEME}/view.html",
    #     ], context)

    @csrf_protect_m
    def changelist_view(self, request, extra_context=None):
        extra_context = extra_context or {}
        extra_context["description"] = self.description
        return super(EntityAdmin, self).changelist_view(request, extra_context)

    # todo check and maybe bring back
    # def get_changelist(self, request, **kwargs):
    #     return RaChangeList

    def save_model(self, request, obj, form, change):
        obj.lastmod_user = request.user
        obj.lastmod_date = now()
        if not change:
            obj.owner = request.user
            obj.creation_date = now()
        super(EntityAdmin, self).save_model(request, obj, form, change)

    def get_urls(self):
        # Override ModelAdmin
        from django.urls import re_path as url

        def wrap(view):
            def wrapper(*args, **kwargs):
                return self.admin_site.admin_view(view)(*args, **kwargs)

            return update_wrapper(wrapper, view)

        info = self.model._meta.app_label, self.model._meta.model_name
        admin_site = self.admin_site
        reversion_urls = [
            url(
                "^recover/$",
                admin_site.admin_view(self.recoverlist_view),
                name="%s_%s_recoverlist" % info,
            ),
            url(
                "^recover/([^/]+)/$",
                admin_site.admin_view(self.recover_view),
                name="%s_%s_recover" % info,
            ),
            url(
                "^([^/]+)/history/([^/]+)/$",
                admin_site.admin_view(self.revision_view),
                name="%s_%s_revision" % info,
            ),
        ]

        urlpatterns = super(EntityAdmin, self).get_urls()

        my_urls = [
            # path('autocomplete/', wrap(self.autocomplete_view), name='%s_%s_autocomplete' % info),
            url(
                r"^slug/(?P<slug>[\w-]+)/$",
                self.admin_site.admin_view(self.get_by_slug),
                name="%s_%s_get-by-slug" % info,
            ),
        ]
        return my_urls + reversion_urls + urlpatterns

    def get_by_slug(self, request, **kwargs):
        query = self.model.objects.filter(**kwargs)
        result = query.exists()
        if result:
            query = query[0]
            admin_url = reverse(
                "%s:%s_%s_change"
                % (
                    app_settings.RA_ADMIN_SITE_NAME,
                    self.model._meta.app_label,
                    self.model._meta.model_name,
                ),
                args=(query.id,),
            )
            return HttpResponseRedirect(admin_url)
        else:
            raise Http404

    def formfield_for_dbfield(self, db_field, request, **kwargs):
        if isinstance(db_field, models.ForeignKey):
            if db_field.__class__ in self.formfield_overrides:
                kwargs = dict(self.formfield_overrides[db_field.__class__], **kwargs)

            formfield = self.formfield_for_foreignkey(db_field, request, **kwargs)
            if formfield and db_field.name not in self.raw_id_fields:
                related_modeladmin = self.admin_site._registry.get(
                    db_field.remote_field.model
                )
                wrapper_kwargs = {}
                if related_modeladmin:
                    wrapper_kwargs.update(
                        can_add_related=related_modeladmin.has_add_permission(request),
                        can_change_related=related_modeladmin.has_change_permission(
                            request
                        ),
                        can_delete_related=related_modeladmin.has_delete_permission(
                            request
                        ),
                    )
                    formfield.widget = RaRelatedFieldWidgetWrapper(
                        formfield.widget,
                        db_field.remote_field,
                        self.admin_site,
                        **wrapper_kwargs,
                    )
            return app_settings.RA_FORMFIELD_FOR_DBFIELD_FUNC(
                self, db_field, formfield, request, **kwargs
            )

        # if db_field.name == 'date' and db_field.__class__ == DateTimeField:
        #     field = forms.SplitDateTimeField(widget=AdminSplitDateTimeNoBr, label=_('Date'))
        # else:
        field = super(EntityAdmin, self).formfield_for_dbfield(
            db_field, request, **kwargs
        )
        if db_field.name == "slug" and self.enable_next_serial:
            field.initial = self.get_next_serial()
        elif db_field.name == "date":
            field.initial = now()
        return app_settings.RA_FORMFIELD_FOR_DBFIELD_FUNC(
            self, db_field, field, request, **kwargs
        )

    def get_next_serial(self):
        from erp_framework.base.helpers import get_next_serial

        return get_next_serial(self.model)

    def whole_changeform_validation(self, request, form, formsets, change, **kwargs):
        """
        A Hook for validating the whole changeform
        :param form: the ModelAdmin Form
        :param formsets: inline formsets
        :param kwargs: extra kwargs
        :return: True for valid [default] False for Invalid
        """
        return True

    def pre_save(self, form, formsets, change):
        """
        Hook for doing any final computation setting before saving
        :param form:
        :param formsets:
        :param change:
        :return:
        """
        pass

    def post_save(self, request, new_object, form, formsets, change):
        """
        Hook for doing any final logic after saving
        :param form:
        :param formsets:
        :param change:
        :return:
        """
        pass

    ###
    def _changeform_view(self, request, object_id, form_url, extra_context):
        """
        A copy of original django method with 4 tricks
        1. Calls whole_changeform_validation
        2. Calls pre_save() before saving
        3. Calls post_save() after saving
        4. emits the signal changeform_saved signal

        :param request:
        :param object_id:
        :param form_url:
        :param extra_context:
        :return:
        """
        to_field = request.POST.get(TO_FIELD_VAR, request.GET.get(TO_FIELD_VAR))
        if to_field and not self.to_field_allowed(request, to_field):
            raise DisallowedModelAdminToField(
                "The field %s cannot be referenced." % to_field
            )

        model = self.model
        opts = model._meta

        if request.method == "POST" and "_saveasnew" in request.POST:
            object_id = None

        add = object_id is None

        if add:
            if not self.has_add_permission(request):
                raise PermissionDenied
            obj = None

        else:
            obj = self.get_object(request, unquote(object_id), to_field)

            if not self.has_view_or_change_permission(request, obj):
                raise PermissionDenied

            if obj is None:
                return self._get_obj_does_not_exist_redirect(request, opts, object_id)

        ModelForm = self.get_form(request, obj, change=not add)
        if request.method == "POST":
            form = ModelForm(request.POST, request.FILES, instance=obj)
            form_validated = form.is_valid()
            if form_validated:
                new_object = self.save_form(request, form, change=not add)
            else:
                new_object = form.instance
            formsets, inline_instances = self._create_formsets(
                request, new_object, change=not add
            )
            if (
                all_valid(formsets)
                and form_validated
                and self.whole_changeform_validation(request, form, formsets, not add)
            ):
                self.pre_save(form, formsets, change=not add)
                self.save_model(request, new_object, form, not add)
                self.save_related(request, form, formsets, not add)
                self.post_save(request, new_object, form, formsets, not add)
                changeform_saved.send(self.model, instance=new_object, created=add)

                change_message = self.construct_change_message(
                    request, form, formsets, add
                )
                if add:
                    self.log_addition(request, new_object, change_message)
                    return self.response_add(request, new_object)
                else:
                    self.log_change(request, new_object, change_message)
                    return self.response_change(request, new_object)
            else:
                form_validated = False
        else:
            if add:
                initial = self.get_changeform_initial_data(request)
                form = ModelForm(initial=initial)
                formsets, inline_instances = self._create_formsets(
                    request, form.instance, change=False
                )
            else:
                form = ModelForm(instance=obj)
                formsets, inline_instances = self._create_formsets(
                    request, obj, change=True
                )

        if not add and not self.has_change_permission(request, obj):
            readonly_fields = flatten_fieldsets(self.get_fieldsets(request, obj))
        else:
            readonly_fields = self.get_readonly_fields(request, obj)
        adminForm = helpers.AdminForm(
            form,
            list(self.get_fieldsets(request, obj)),
            # Clear prepopulated fields on a view-only form to avoid a crash.
            self.get_prepopulated_fields(request, obj)
            if add or self.has_change_permission(request, obj)
            else {},
            readonly_fields,
            model_admin=self,
        )
        media = self.media + adminForm.media

        inline_formsets = self.get_inline_formsets(
            request, formsets, inline_instances, obj
        )
        for inline_formset in inline_formsets:
            media = media + inline_formset.media

        if add:
            name = _("Add %s")
        elif self.has_change_permission(request, obj):
            name = _("Change %s")
        else:
            name = _("View %s")
        context = {
            **self.admin_site.each_context(request),
            "name": name % opts.verbose_name,
            "adminform": adminForm,
            "object_id": object_id,
            "original": obj,
            "is_popup": IS_POPUP_VAR in request.POST or IS_POPUP_VAR in request.GET,
            "to_field": to_field,
            "media": media,
            "inline_admin_formsets": inline_formsets,
            "errors": helpers.AdminErrorList(form, formsets),
            "preserved_filters": self.get_preserved_filters(request),
        }

        # Hide the "Save" and "Save and continue" buttons if "Save as New" was
        # previously chosen to prevent the interface from getting confusing.
        if (
            request.method == "POST"
            and not form_validated
            and "_saveasnew" in request.POST
        ):
            context["show_save"] = False
            context["show_save_and_continue"] = False
            # Use the change template instead of the add template.
            add = False

        context.update(extra_context or {})

        return self.render_change_form(
            request, context, add=add, change=not add, obj=obj, form_url=form_url
        )

    def get_actions(self, request):
        actions = super(EntityAdmin, self).get_actions(request)
        if not (app_settings.RA_ENABLE_ADMIN_DELETE_ALL and request.user.is_superuser):
            if "delete_selected" in actions:
                del actions["delete_selected"]
        return actions

    def get_list_display(self, request):
        list_display = super(EntityAdmin, self).get_list_display(request)

        if self.has_view_permission(request):
            list_display = ("get_stats_icon",) + list_display

        return list_display


class TransactionAdmin(EntityAdmin):
    enable_view_view = False
    list_per_page = 50
    view_on_site = False
    list_display = ("slug", "date", "value", "get_history_link")
    list_display_links = ("slug", "date")
    fields = (("slug", "date"),)
    exclude = ("type",)
    formfield_overrides = {
        # DateTimeField: {'widget': AdminSplitDateTimeNoBr},
        # DecimalField: {'widget': NumberInput},
        # ForeignKey: {'widget': RaBootstrapForeignKeyWidget},
    }
    date_hierarchy = "date"
    type = None

    # Copy values from parent model to child inline models,
    # If `all-fk` then automatically it will copy all foreign keys included in the fields of the add/change Form
    # other options are None , or a list of field names
    copy_to_inlines = "all-fk"

    copy_form_notes_to_formset = False
    search_fields = ["date", "slug"]

    def formfield_for_dbfield(self, db_field, request, **kwargs):
        form_field = super(TransactionAdmin, self).formfield_for_dbfield(
            db_field, request, **kwargs
        )
        form_field = app_settings.RA_FORMFIELD_FOR_DBFIELD_FUNC(
            self, db_field, form_field, request, **kwargs
        )
        return form_field

    def get_foreign_keys(self, form_fields=None):
        fks = [
            field.name
            for field in self.model._meta.get_fields(include_parents=False)
            if field.get_internal_type() == "ForeignKey"
        ]
        if form_fields:
            fks = set(fks).intersection(set(form_fields))
        return fks

    def get_copy_to_inlines(self, form, formset):
        if self.copy_to_inlines == "all-fk":
            fks = self.get_foreign_keys(form.fields.keys())
            return fks
        return self.copy_to_inlines or []

    def get_list_filter(self, request):
        list_filter = super().get_list_filter(request)
        if not list_filter:
            return self.get_foreign_keys(flatten_list(self.get_fields(request)))
        return list_filter

    def save_formset(self, request, form, formset, change):
        copy_to_inlines = self.get_copy_to_inlines(form, formset)
        for formset_form in formset:
            formset_form.instance.date = form.instance.date
            formset_form.instance.type = form.instance.type
            formset_form.instance.slug = form.instance.slug
            formset_form.instance.lastmod_user = request.user
            if self.copy_form_notes_to_formset:
                formset_form.instance.notes = form.instance.notes
            for copy_item in copy_to_inlines:
                formset_form.instance.__setattr__(
                    copy_item, form.instance.__getattribute__(copy_item)
                )
                # print(copy_item, form.instance.__getattribute__(copy_item))
            if formset_form.instance.pk is not None:
                formset_form.instance.save()
        formset.save()


class TransactionItemAdmin(admin.TabularInline):
    template = f"{app_settings.RA_THEME}/edit_inline/tabular.html"
    extra = 2
    exclude = (
        "slug",
        "date",
        "type",
        "lastmod",
        "lastmod_user",
        "owner",
        "creation_date",
    )
    formfield_overrides = {
        models.TextField: {"widget": TextInput},
        DecimalField: {"widget": NumberInput},
        # ForeignKey: {'widget': RaBootstrapForeignKeyWidget},
    }
    view_on_site = False

    """
    To simplify complex forms with inline , making inline permission reflect base form permission
    permission_override_model can be True, False , str Or ModelBase
    """
    permission_override_model = True

    def get_ra_permission_codename(self, action, model_name):
        """
        Returns the codename of the permission for the specified action.
        """
        return "%s_%s" % (action, model_name)

    def get_permission_override_model(self, request, **kwargs):
        """
        Return a string reprsentation of the model to look into its permissions,
        :param request:
        :param kwargs:
        :return:
        """
        if self.permission_override_model is True:
            return self.parent_model._meta.model_name

        elif type(self.permission_override_model) is str:
            return self.permission_override_model

        elif type(self.permission_override_model) is ModelBase:
            return self.permission_override_model._meta.model_name

        else:
            raise ImproperlyConfigured(
                "self.permission_override_model can be True, False , str or ModelBase .Got %s instead "
                % type(self.permission_override_model)
            )

    def has_add_permission(self, request, obj=None):
        if self.opts.auto_created:
            # We're checking the rights to an auto-created intermediate model,
            # which doesn't have its own individual permissions. The user needs
            # to have the change permission for the related model in order to
            # be able to do anything with the intermediate model.
            return self.has_change_permission(request)
        opts = self.opts
        if self.permission_override_model:
            model_name = self.get_permission_override_model(request)
            codename = self.get_ra_permission_codename("add", model_name)
        else:
            codename = get_permission_codename("add", opts)
        return request.user.has_perm("%s.%s" % (opts.app_label, codename))

    def has_change_permission(self, request, obj=None):
        opts = self.opts
        if opts.auto_created:
            # The model was auto-created as intermediary for a
            # ManyToMany-relationship, find the target model
            for field in opts.fields:
                if field.remote_field and field.remote_field.model != self.parent_model:
                    opts = field.remote_field.model._meta
                    break
        if self.permission_override_model:
            model_name = self.get_permission_override_model(request)
            codename = self.get_ra_permission_codename("change", model_name)
        else:
            codename = get_permission_codename("change", opts)
        return request.user.has_perm("%s.%s" % (opts.app_label, codename))

    def has_delete_permission(self, request, obj=None):
        if self.opts.auto_created:
            # We're checking the rights to an auto-created intermediate model,
            # which doesn't have its own individual permissions. The user needs
            # to have the change permission for the related model in order to
            # be able to do anything with the intermediate model.
            return self.has_change_permission(request, obj)
        opts = self.opts
        if self.permission_override_model:
            model_name = self.get_permission_override_model(request)
            codename = self.get_ra_permission_codename("delete", model_name)
        else:
            codename = get_permission_codename("delete", opts)
        return request.user.has_perm("%s.%s" % (opts.app_label, codename))

    def formfield_for_dbfield(self, db_field, request, **kwargs):
        if isinstance(db_field, models.ForeignKey):
            if db_field.__class__ in self.formfield_overrides:
                kwargs = dict(self.formfield_overrides[db_field.__class__], **kwargs)

            formfield = self.formfield_for_foreignkey(db_field, request, **kwargs)
            if formfield and db_field.name not in self.raw_id_fields:
                related_modeladmin = self.admin_site._registry.get(
                    db_field.remote_field.model
                )
                wrapper_kwargs = {}
                if related_modeladmin:
                    wrapper_kwargs.update(
                        can_add_related=related_modeladmin.has_add_permission(request),
                        can_change_related=related_modeladmin.has_change_permission(
                            request
                        ),
                        can_delete_related=related_modeladmin.has_delete_permission(
                            request
                        ),
                    )
                    formfield.widget = RaRelatedFieldWidgetWrapper(
                        formfield.widget,
                        db_field.remote_field,
                        self.admin_site,
                        **wrapper_kwargs,
                    )
        form_field = super(TransactionItemAdmin, self).formfield_for_dbfield(
            db_field, request, **kwargs
        )
        form_field = app_settings.RA_FORMFIELD_FOR_DBFIELD_FUNC(
            self, db_field, form_field, request, **kwargs
        )
        return form_field


class RaGenericTabularInline(GenericTabularInline):
    """
    Implementation of teh needed methods for Generic Tabular inline with RA
    """

    extra = 1
    exclude = (
        "slug",
        "date",
        "type",
        "lastmod",
        "lastmod_user",
        "owner",
        "creation_date",
    )
    formfield_overrides = {
        TextField: {"widget": TextInput},
        DecimalField: {"widget": NumberInput},
    }
    view_on_site = False

    def formfield_for_dbfield(self, db_field, request, **kwargs):
        if isinstance(db_field, models.ForeignKey):
            if db_field.__class__ in self.formfield_overrides:
                kwargs = dict(self.formfield_overrides[db_field.__class__], **kwargs)

            formfield = self.formfield_for_foreignkey(db_field, request, **kwargs)
            if formfield and db_field.name not in self.raw_id_fields:
                related_modeladmin = self.admin_site._registry.get(
                    db_field.remote_field.model
                )
                wrapper_kwargs = {}
                if related_modeladmin:
                    wrapper_kwargs.update(
                        can_add_related=related_modeladmin.has_add_permission(request),
                        can_change_related=related_modeladmin.has_change_permission(
                            request
                        ),
                        can_delete_related=related_modeladmin.has_delete_permission(
                            request
                        ),
                    )
                    formfield.widget = RaRelatedFieldWidgetWrapper(
                        formfield.widget,
                        db_field.remote_field,
                        self.admin_site,
                        **wrapper_kwargs,
                    )
        form_field = super(RaGenericTabularInline, self).formfield_for_dbfield(
            db_field, request, **kwargs
        )
        form_field = app_settings.RA_FORMFIELD_FOR_DBFIELD_FUNC(
            self, db_field, form_field, request, **kwargs
        )
        return form_field


class PrepopulatedAdmin(object):
    prepopulation_querysets = None
    prepopulation_fields = None
    prepopulation_fields_from = None
    prepopulated_inlines = []  # a subset of inlines that will be prepopulated
    prepopulation_extra_data = None  # to do A list of fields

    def add_view(self, request, form_url="", extra_context=None):
        extra_context = extra_context or {}
        extra_context["prepopulation_fields"] = self.prepopulation_fields
        return super(PrepopulatedAdmin, self).add_view(request, form_url, extra_context)

    def change_view(self, request, object_id, form_url="", extra_context=None):
        extra_context = extra_context or {}
        extra_context["prepopulation_fields"] = self.prepopulation_fields
        return super(PrepopulatedAdmin, self).change_view(
            request, object_id, form_url, extra_context
        )

    def get_prepopulattion_queryset(self, inline):
        return self.prepopulation_querysets.get(inline.__class__, None)

    def get_prepopulation_queryset(self, inline, inline_queryset, request, obj, change):
        """
        return a queryset Or None
        """
        try:
            return self.prepopulation_querysets.get(inline.__class__, None)
        except AttributeError:
            raise ImproperlyConfigured(
                "Please check %s.prepopulation_querysets to be a dict type"
                % self.__class__
            )

    def filter_inline_queryset(self, inline_queryset, request, obj, change):
        f = self._get_inline_queryset_filter(inline_queryset)
        if f:
            inline_queryset = inline_queryset.filter(**{f: obj})
        return inline_queryset

    def _get_inline_queryset_filter(self, inline_queryset):
        fields = inline_queryset.model._meta.get_fields()
        for f in fields:
            if type(f) is ForeignKey:
                if f.related_model in self.model.__mro__:
                    return f.name
        return ""

    def get_prepopulation_extra_data(self, inline, data):
        """
        Hook to add Extra data (other then prepoputation field) to the prepopulated form
        """
        return {}

    def get_formset_initial(self, inline, inline_queryset, request, obj, change):
        self.prepopulation_fields_from = self.prepopulation_fields_from or {}

        prepopulateion_queryset = self.get_prepopulation_queryset(
            inline, inline_queryset, request, obj, change
        )
        prepopulation_field = self.prepopulation_fields.get(inline.__class__, None)
        prepopulation_field_from = self.prepopulation_fields_from.get(
            inline.__class__, "id"
        )
        prepopulation_field_from = (
            prepopulation_field_from
            if prepopulation_field_from == "id"
            else "%s_id" % prepopulation_field_from
        )

        if prepopulateion_queryset is None:
            return []
        else:
            assert prepopulation_field is not None
            initial_data = prepopulateion_queryset.values(prepopulation_field_from)

            id_list = [x[prepopulation_field_from] for x in initial_data]
            initial_data = [x for x in initial_data]
            return_data = []
            if change:
                filtered_queryset = self.filter_inline_queryset(
                    inline_queryset, request, obj, change
                )
                for item in filtered_queryset:
                    try:
                        x = id_list.index(
                            item.__getattribute__("%s_id" % prepopulation_field)
                        )
                        id_list.pop(x)
                        initial_data.pop(x)
                    except ValueError:
                        pass
                for i, data in enumerate(initial_data):
                    return_data.append(
                        {prepopulation_field: initial_data[i][prepopulation_field_from]}
                    )
                request._initial_data = return_data
                return return_data

            elif not change:
                ret_val = []
                for item in initial_data:
                    ret_val.append(
                        {prepopulation_field: item[prepopulation_field_from]}
                    )
                return ret_val

        return []

    def _create_formsets(self, request, obj, change):
        "Helper function to generate formsets for add/change_view."
        formsets = []
        inline_instances = []
        prefixes = {}
        get_formsets_args = [request]
        if change:
            get_formsets_args.append(obj)
        for FormSet, inline in self.get_formsets_with_inlines(*get_formsets_args):
            prefix = FormSet.get_default_prefix()
            prefixes[prefix] = prefixes.get(prefix, 0) + 1
            if prefixes[prefix] != 1 or not prefix:
                prefix = "%s-%s" % (prefix, prefixes[prefix])
            inline_queryset = inline.get_queryset(request)
            formset_params = {
                "instance": obj,
                "prefix": prefix,
                "queryset": inline_queryset,
            }
            if request.method == "POST":
                formset_params.update(
                    {
                        "data": request.POST,
                        "files": request.FILES,
                        "save_as_new": "_saveasnew" in request.POST,
                    }
                )
            else:
                initial_data = self.get_formset_initial(
                    inline, inline_queryset, request, obj, change
                )
                # Not here is a necessity, in case you find unuseful revise again
                if (
                    inline.__class__ in self.prepopulated_inlines
                    or not self.prepopulated_inlines
                ):
                    FormSet.extra = len(initial_data)
                if initial_data:
                    formset_params.update(
                        {
                            "initial": initial_data,
                        }
                    )
            formsets.append(FormSet(**formset_params))
            inline_instances.append(inline)
        return formsets, inline_instances


class RaPrePopulatedAdmin(PrepopulatedAdmin, EntityAdmin):
    pass


class RaMovementPrepopulatedAdmin(RaPrePopulatedAdmin):
    enable_view_view = False
    list_per_page = 50
    copy_to_inlines = []
    copy_form_notes_to_formset = False

    def save_formset(self, request, form, formset, change):
        for formset_form in formset:
            formset_form.instance.date = form.instance.date
            formset_form.instance.type = form.instance.type
            formset_form.instance.slug = form.instance.slug
            formset_form.instance.lastmod_user = request.user
            if self.copy_form_notes_to_formset:
                formset_form.instance.notes = form.instance.notes
            for copy_item in self.copy_to_inlines:
                formset_form.instance.__setattr__(
                    copy_item, form.instance.__getattribute__(copy_item)
                )

        formset.save()


class RaUserAdmin(RaThemeMixin, UserTabularPermissionsMixin, UserAdmin):
    enable_view_view = False
    date_hierarchy = None
    fields = None
    list_display_links = ["username"]
    change_user_password_template = (
        f"{app_settings.RA_THEME}/auth/user/change_password.html"
    )
    list_display = ("username", "is_superuser", "last_login", "is_active")

    save_on_top = True

    fieldsets = (
        (None, {"fields": ("username", "password")}),
        (_("Personal info"), {"fields": (("first_name", "last_name", "email"),)}),
        (
            _("Permissions"),
            {"fields": (("is_active", "is_superuser"), "groups", "user_permissions")},
        ),
        # (_('Important dates'), {'fields': (('last_login', 'date_joined'),)}),
    )
    list_filter = ("is_superuser", "is_active", "groups")
    form = RaUserChangeForm

    def get_actions(self, request):
        actions = super(RaUserAdmin, self).get_actions(request)
        if "delete_selected" in actions:
            del actions["delete_selected"]
        return actions

    def has_delete_permission(self, request, obj=None):
        return False

    def save_model(self, request, obj, form, change):
        obj.is_staff = True
        super(RaUserAdmin, self).save_model(request, obj, form, change)

    def get_inline_instances(self, request, obj=None):
        if obj is None:
            return []
        return super(RaUserAdmin, self).get_inline_instances(request, obj)

    def changelist_view(self, request, extra_context=None):
        extra_context = {"has_detached_sidebar": True}
        return super(RaUserAdmin, self).changelist_view(request, extra_context)


class RaGroupAdmin(RaThemeMixin, GroupTabularPermissionsMixin, GroupAdmin):
    enable_view_view = False
    date_hierarchy = None
    fields = None
    list_display_links = ["name"]
    list_display = ["name"]

    save_on_top = True


erp_admin_site.register(Group, RaGroupAdmin)
erp_admin_site.register(User, RaUserAdmin)



from django.contrib.admin.models import LogEntry
from django.utils.translation import gettext_lazy as _


class MyActivity(LogEntry):
    class Meta:
        proxy = True
        verbose_name = _("My Activity")
        verbose_name_plural = _("My Activities")



import copy
import hashlib
import logging

import simplejson as json
from django.conf import settings
from django.contrib.auth.mixins import AccessMixin, UserPassesTestMixin
from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured, PermissionDenied
from django.db.models import Q
from django.http import JsonResponse, HttpResponse, HttpResponseRedirect
from django.template.defaultfilters import capfirst
from django.urls import reverse, reverse_lazy
from django.utils.translation import get_language_bidi, gettext
from django.views.generic import TemplateView

from erp_framework.base import app_settings, registry
from erp_framework.base.app_settings import RA_ADMIN_SITE_NAME
from erp_framework.base.helpers import dictsort
from erp_framework.reporting.forms import OrderByForm
from erp_framework.reporting.printing import (
    regroup_data,
    HTMLPrintingClass,
    ExportToCSV,
)
from slick_reporting.form_factory import report_form_factory
from slick_reporting.generator import ReportGenerator
from slick_reporting.views import SlickReportViewBase

# from .generator import ReportGenerator
# from .meta_data import ReportMetaData

logger = logging.getLogger("erp_framework.reporting")


class RaMultiplePermissionsRequiredMixin(AccessMixin):
    """
    View mixin which allows you to specify two types of permission
    requirements. The `permissions` attribute must be a dict which
    specifies two keys, `all` and `any`. You can use either one on
    its own or combine them. The value of each key is required to be a
    list or tuple of permissions. The standard Django permissions
    style is not strictly enforced. If you have created your own
    permissions in a different format, they should still work.

    By specifying the `all` key, the user must have all of
    the permissions in the passed in list.

    By specifying The `any` key , the user must have ONE of the set
    permissions in the list.

    Class Settings
        `permissions` - This is required to be a dict with one or both
            keys of `all` and/or `any` containing a list or tuple of
            permissions.
        `login_url` - the login url of site
        `redirect_field_name` - defaults to "next"
        `raise_exception` - defaults to False - raise 403 if set to True

    Example Usage
        class SomeView(MultiplePermissionsRequiredMixin, ListView):
            ...
            #required
            permissions = {
                "all": ("blog.add_post", "blog.change_post"),
                "any": ("blog.delete_post", "user.change_user")
            }

            #optional
            login_url = "/signup/"
            redirect_field_name = "hollaback"
            raise_exception = True
    """

    permissions = None  # Default required perms to none
    permission_or_test = False
    login_url = reverse_lazy("%s:login" % RA_ADMIN_SITE_NAME)
    _bypass = False

    def get_or_test(self):
        return False

    def get_permissions(self):
        return self.permissions

    def return_access_denied(self):
        raise PermissionDenied
        # return HttpResponseForbidden()
        # return HttpResponseRedirect(reverse('access_denied'))

    def dispatch(self, request, *args, **kwargs):
        from django.contrib.auth.views import redirect_to_login

        if request.user.is_anonymous:
            return redirect_to_login(
                request.get_full_path(),
                self.get_login_url(),
                self.get_redirect_field_name(),
            )

        self._check_permissions_attr()

        perms_all = self.get_permissions().get("all") or None
        perms_any = self.get_permissions().get("any") or None

        self._check_permissions_keys_set(perms_all, perms_any)
        self._check_perms_keys("all", perms_all)
        self._check_perms_keys("any", perms_any)

        # If perms_all, check that user has all permissions in the list/tuple
        if perms_all:
            if not request.user.has_perms(perms_all):
                if not (self.permission_or_test and self.get_or_test()):
                    # if self.raise_exception:
                    raise PermissionDenied
                    # return access_denied

        # If perms_any, check that user has at least one in the list/tuple
        if perms_any:
            has_one_perm = False
            for perm in perms_any:
                if request.user.has_perm(perm):
                    has_one_perm = True
                    break

            if not has_one_perm:
                raise PermissionDenied

        return super(RaMultiplePermissionsRequiredMixin, self).dispatch(
            request, *args, **kwargs
        )

    def _check_permissions_attr(self):
        """
        Check permissions attribute is set and that it is a dict.
        """
        permissions = self.get_permissions()
        if permissions is None or not isinstance(permissions, dict):
            raise ImproperlyConfigured(
                "'PermissionsRequiredMixin' requires "
                "'permissions' attribute to be set to a dict."
            )

    def _check_permissions_keys_set(self, perms_all=None, perms_any=None):
        """
        Check to make sure the keys `any` or `all` are not both blank.
        If both are blank either an empty dict came in or the wrong keys
        came in. Both are invalid and should raise an exception.
        """
        if perms_all is None and perms_any is None and not self._bypass:
            raise ImproperlyConfigured(
                "'PermissionsRequiredMixin' requires"
                "'permissions' attribute to be set to a dict and the 'any' "
                "or 'all' key to be set."
            )

    def _check_perms_keys(self, key=None, perms=None):
        """
        If the permissions list/tuple passed in is set, check to make
        sure that it is of the type list or tuple.
        """
        if perms and not isinstance(perms, (list, tuple)):
            raise ImproperlyConfigured(
                "'MultiplePermissionsRequiredMixin' "
                "requires permissions dict '%s' value to be a list "
                "or tuple." % key
            )


class ReportListBase(RaMultiplePermissionsRequiredMixin, TemplateView):
    """
    Base class to create a report list page
    """

    def get_meta_data(self):
        """
        Gets Meta data used for Page title , breadcrumbs links etc.
        Make sure that opts is a model._meta or s sufficient dictionary
        :return: tuple (verbose_name, verbose_name_plural, page_title, model_meta)
        """
        raise NotImplemented

    def get_permissions(self):
        """
        Override of RAAccessControl.get_permissions
        :return: a dictionary with 'any' and/ or 'all' permission. required
        """
        raise NotImplemented

    def get_reports_map(self):
        """
        Hook to get reports. By default it uses the helper `get_reports_map` to get
         the reports based on the base_model, However you can override and return your list of reports

        :return: a dictionary with two values `slugs` and `reports`
        slugs: map to a list of report slugs
        reports: map to a list of ReportView classes
        """
        raise NotImplemented


class ReportList(ReportListBase):
    template_name = f"erp_framework/report_list.html"
    _bypass = True

    def get_order_list(self):
        from erp_framework.admin.admin import erp_admin_site

        model_admin = erp_admin_site.get_admin_by_model_name(self.kwargs["base_model"])
        if model_admin:
            try:
                return model_admin["admin"].typed_reports_order_list or []
            except AttributeError:
                # The admin class does not have an order list for teh reports
                pass
        return []

    def get_permissions(self):
        return {}

    def get_reports_map(self):
        from erp_framework.admin.admin import get_reports_map
        from erp_framework.base.registry import get_ra_model_by_name

        model = get_ra_model_by_name(self.kwargs["base_model"])
        try:
            model_name = model.get_class_name().lower()
        except:
            model_name = model._meta.model_name
        val = get_reports_map(
            model_name, self.request.user, self.request, self.get_order_list()
        )
        return val

    def get_meta_data(self):
        model = registry.get_ra_model_by_name(self.kwargs["base_model"])
        verbose_name = model._meta.verbose_name
        verbose_name_plural = model._meta.verbose_name_plural
        is_bidi = get_language_bidi()
        if is_bidi:
            page_title = "%s %s" % (gettext("reports"), model._meta.verbose_name_plural)
        else:
            page_title = "%s %s" % (model._meta.verbose_name_plural, gettext("reports"))
        opts = model._meta
        return verbose_name, verbose_name_plural, page_title, opts

    def get_context_data(self, **kwargs):
        from erp_framework.admin.admin import erp_admin_site

        context = super(ReportList, self).get_context_data(**kwargs)

        context["reports"] = self.get_reports_map()

        v, vp, page_title, opts = self.get_meta_data()
        context["verbose_name"] = v
        context["verbose_name_plural"] = vp
        context["page_title"] = page_title
        context["title"] = page_title

        context["opts"] = opts

        context["RA_ADMIN_SITE_NAME"] = app_settings.RA_ADMIN_SITE_NAME
        context["has_detached_sidebar"] = True
        context["RA_ADMIN_SITE_NAME"] = app_settings.RA_ADMIN_SITE_NAME
        context["is_report"] = True
        context["base_model"] = self.kwargs["base_model"].lower()
        context["report_slug"] = ""

        extra_context = erp_admin_site.each_context(self.request)
        context.update(extra_context)
        return context


class ReportView(UserPassesTestMixin, SlickReportViewBase):
    """
    The Base class for reports .
    It handles the report ajax request, load the report form which provides the needed filers,
    Load the report generator class , csv_export_class and the printing class.

    """

    # class responsible for generating the report, applying the filter and computing
    report_generator_class = ReportGenerator

    # class responsible for supplying meta data for front end
    # report_meta_data_class = ReportMetaData

    # class responsible for customizing the output for print
    printing_class = HTMLPrintingClass

    # required. the model which holds the data
    report_model = None

    # required , the main model which the report is about.
    # We will be getting the doc_types (if any) from this model and be basing the calculation on.
    base_model = None

    # the report form, A subclass of ReportForm is to expected
    form_class = None

    # control the settings of report
    form_settings = None

    # control the chart settings, passed to front end as is.
    chart_settings = None

    other_namespaces = None
    columns = None

    report_slug = ""
    page_title = None
    report_title = ""
    date_field = "date"
    # default order by for the results.
    # ordering can also be controlled on run time by passing order_by='field_name'
    # For DESC order supply order_by='-field_name'
    default_order_by = None

    # this report will not be visible on the menu or accessed on its own
    hidden = False

    # will swap the sign on the report, useful when reporting on object which main side is credit

    # Control the header report function
    must_exist_filter = None
    header_report = None

    # to limit records not to exceed certain number, useful for very large reports
    limit_records = False

    perform_regroup_when_print = False  # Handling printing header & sub data

    # size for printing , useful on very large data
    print_buffer = 10000

    # control the caching
    cache = True
    cache_duration = 300
    with_type = True

    admin_site_name = "erp_admin"
    template_name = "erp_framework/report.html"

    def get_context_data(self, **kwargs):
        from erp_framework.admin.admin import erp_admin_site

        context = super().get_context_data(**kwargs)
        extra_context = erp_admin_site.each_context(self.request)
        context.update(extra_context)
        context["is_report"] = True
        context["current_base_model_name"] = self.base_model.__name__.lower()
        context["base_model"] = self.base_model
        context["report_slug"] = self.get_report_slug()
        context["report"] = self

        return context

    def get_doc_types_q_filters(self):
        return [], []

        # doc_types = registry.get_model_doc_type_map(self.base_model.__name__)
        # return [Q(type__in=doc_types['plus_list'])], [Q(type__in=doc_types['minus_list'])]

    def get_report_generator(self, queryset, for_print):
        q_filters, kw_filters = self.form.get_filters()
        if self.crosstab_model:
            self.crosstab_ids = self.form.get_crosstab_ids()

        crosstab_compute_reminder = (
            self.form.get_crosstab_compute_reminder()
            if self.request.GET or self.request.POST
            else self.crosstab_compute_reminder
        )
        doc_type_plus_list, doc_type_minus_list = [], []

        if self.with_type:
            doc_type_plus_list, doc_type_minus_list = self.get_doc_types_q_filters()

        time_series_pattern = self.time_series_pattern
        if self.time_series_selector:
            time_series_pattern = self.form.cleaned_data["time_series_pattern"]

        return self.report_generator_class(
            self.get_report_model(),
            start_date=self.form.cleaned_data["start_date"],
            end_date=self.form.cleaned_data["end_date"],
            q_filters=q_filters,
            kwargs_filters=kw_filters,
            date_field=self.date_field,
            main_queryset=queryset,
            print_flag=for_print,
            limit_records=self.limit_records,
            swap_sign=self.swap_sign,
            columns=self.columns,
            group_by=self.group_by,
            time_series_pattern=time_series_pattern,
            time_series_columns=self.time_series_columns,
            crosstab_model=self.crosstab_model,
            crosstab_ids=self.crosstab_ids,
            crosstab_columns=self.crosstab_columns,
            crosstab_compute_reminder=crosstab_compute_reminder,
            format_row_func=self.format_row,
            doc_type_plus_list=doc_type_plus_list,
            doc_type_minus_list=doc_type_minus_list,
        )

    @classmethod
    def get_report_slug(cls):
        return cls.report_slug or cls.__name__.lower()

    @classmethod
    def get_base_model_name(cls):
        """
        A convenience method to get the base model name
        :return:
        """
        return cls.base_model._meta.model_name

    @classmethod
    def get_url(cls):
        url = reverse(
            f"{RA_ADMIN_SITE_NAME}:report_list", args=(cls.get_base_model_name(),)
        )
        url += "%s/" % cls.get_report_slug()
        return url

    def test_func(self):
        app_label = self.base_model._meta.app_label
        codename = f"view_{self.get_report_slug()}"
        return self.request.user.has_perm(f"{app_label}.{codename}")

    def handle_no_permission(self):
        if self.request.user.is_authenticated:
            if self.request.META.get("HTTP_X_REQUESTED_WITH") == "XMLHttpRequest":
                return JsonResponse({}, status=403)
            else:
                raise PermissionDenied
            #     return HttpResponseRedirect(reverse('erp_admin::access-denied'))
        # else:
        return HttpResponseRedirect(reverse("erp_admin:login"))

    @classmethod
    def get_absolute_url(self):
        return reverse(
            "admin:report",
            args=(self.base_model._meta.model_name, self.get_report_slug()),
            current_app=self.admin_site_name,
        )

    @classmethod
    def get_report_model(cls):
        """
        Problem: During tests, override settings is used, making the report model always returning the model
        'first to be found' not the potentially swapped one ,raising an error. so , it is advised to use this method instead
            of declaring the report model on the module level.
        :return: the Model to use
        """
        return cls.report_model

    @staticmethod
    def form_filter_func(fkeys_dict):
        output = {}
        for k, v in fkeys_dict.items():
            if k not in ["owner_id", "polymorphic_ctype_id", "lastmod_user_id"]:
                output[k] = v
        return output

    @classmethod
    def get_form_class(cls):
        """
        As Report Model might be swapped, form_class cant be accurately generated on model load,
        hence this function.
        :return: form_class
        """
        return cls.form_class or report_form_factory(
            cls.get_report_model(),
            crosstab_model=cls.crosstab_model,
            display_compute_reminder=cls.crosstab_compute_reminder,
            fkeys_filter_func=cls.form_filter_func,
            initial=cls.get_form_initial(),
            show_time_series_selector=cls.time_series_selector,
            time_series_selector_choices=cls.time_series_selector_choices,
            time_series_selector_default=cls.time_series_selector_default,
            time_series_selector_allow_empty=cls.time_series_selector_allow_empty,
        )

    def dispatch(self, request, *args, **kwargs):
        report_slug = kwargs.get("report_slug", False)
        if report_slug:
            self.report_slug = report_slug
        val = super(ReportView, self).dispatch(request, *args, **kwargs)

        return val

    def get_printing_class(
        self,
    ):
        return self.printing_class

    def is_print_request(self):
        return self.request.GET.get("print", False)

    @classmethod
    def get_initialized_form(cls):
        """
        Get the form_class initialized.
        :return:
        """
        form_class = cls.get_form_class()
        return form_class()

    @classmethod
    def get_all_print_settings(cls):
        # todo review
        return {}

    def return_header_report_or_none(self):
        """
        Check if we need to return o
        :return:
        """

        if self.must_exist_filter:
            filter_exist = self.request.GET.get(self.must_exist_filter, "")
            # if the filter exists but the the user choose several entities
            filter_exist = filter_exist.find(",") == -1 and bool(filter_exist)
            if not filter_exist:
                return JsonResponse({}, status=204)
                return self.header_report.as_view()(
                    self.request,
                    original_report_slug=self.get_report_slug(),
                    as_header=True,
                )
        return None

    def get_print_response(self, response, header_report, must_exist_filter):
        """
        Invoke the printing class and return the printing response
        :return:  Response
        """

        if must_exist_filter and header_report:
            if self.perform_regroup_when_print:
                response["sub_data"] = regroup_data(
                    response["data"], header_report, must_exist_filter
                )
                response["header_group"] = header_report
            else:
                _available_buffer = self.print_buffer
                must_exist_filter = must_exist_filter.replace("__slug", "_id")
                response["sub_data"] = {}
                json_data = json.loads(header_report.content)
                response["header_data"] = []
                response["header_group"] = must_exist_filter

                for line in json_data["data"]:
                    pivot_id = line[must_exist_filter]
                    new_request = copy.copy(self.request)
                    get_data = new_request.GET.copy()
                    get_data[must_exist_filter] = pivot_id
                    get_data["print_recursive"] = True
                    get_data["get_group"] = False
                    new_request.GET = get_data
                    pivot_report = self.__class__.as_view()(new_request)
                    data_check = pivot_report.get("data", [])
                    if not len(data_check) > 0:
                        _available_buffer -= len(pivot_report.get("data", []))
                        response["sub_data"][pivot_id] = pivot_report
                        response["header_data"].append(line)
                        if _available_buffer < 0:
                            break

        return self.get_printing_class()(
            self.request,
            self.form_settings or {},
            response,
            header_report,
            must_exist_filter,
            self,
        ).get_response()

    # todo delete
    def cache_and_return(self, results):
        return results
        status, key = self.get_cache_status_and_key(self.request)
        if key:
            cache.set(key, results, self.cache_duration)
        return results

    def cache_view(self):
        return self.cache and app_settings.RA_CACHE_REPORTS

    def get_cache_status_and_key(self, request):
        if not self.cache or not app_settings.RA_CACHE_REPORTS:
            return False, None

        url = request.get_full_path()
        url_parts = url.split("&")
        url_parts = [x for x in url_parts if not x.startswith("csrf")]
        no_cache_request = "no-cache" in url_parts
        if no_cache_request:
            url_parts.remove("no-cache")
        url = "".join(url_parts)
        m = hashlib.md5()
        if app_settings.RA_CACHE_REPORTS_PER_USER:
            m.update(f"{request.user.pk}".encode("utf-8"))
        m.update(url.encode("utf-8"))
        cache_key = m.hexdigest()
        return not no_cache_request, cache_key

    def get_export_response(self, response, header_report, must_exist_filter):
        return self.csv_export_class(
            self.request,
            self.form_settings or {},
            response,
            header_report,
            must_exist_filter,
            self,
        ).get_response()

    def _get(self, request, *args, **kwargs):
        cache_status, cache_key = self.get_cache_status_and_key(request)
        if cache_status and cache_key:
            cache_val = cache.get(cache_key)
            if cache_val:
                return cache_val

        # self.parse()

        form_class = self.get_form_class()
        self.form = self.get_form(form_class)
        enable_print = self.is_print_request()
        export_csv = request.GET.get("csv", False)

        # if request.META.get('HTTP_X_REQUESTED_WITH') == 'XMLHttpRequest' or enable_print or export_csv or (
        #         request.GET.get('ajax') == 'true' and settings.DEBUG):
        if request.GET and self.form.is_valid():
            must_exist_filter = self.must_exist_filter or ""
            must_exist_filter = must_exist_filter.replace("_id", "__slug")
            header_report = self.return_header_report_or_none()
            # Case of a report that needs a specific filter to run,
            # in case it doesn't exist, we return the `header_report`
            if header_report and not enable_print:
                return self.cache_and_return(header_report)
            as_header = self.kwargs.get("as_header", False)

            response = self.get_report_results(enable_print or export_csv)

            if (export_csv or enable_print) and not as_header:
                response = self.prepare_results_for_printing(response)
                if export_csv:
                    return self.get_export_response(
                        response, header_report, must_exist_filter
                    )

                return self.get_print_response(
                    response, header_report, must_exist_filter
                )
            return self.cache_and_return(
                HttpResponse(
                    self.serialize_to_json(response), content_type="application/json"
                )
            )
        else:
            return self.form_invalid(self.form)
        # else:
        #     # Accessing the report page directly is not allowed
        #     return HttpResponseRedirect(
        #         reverse(f'{RA_ADMIN_SITE_NAME}:report_list', args=(self.get_base_model_name(),)))

    @classmethod
    def get_report_title(cls):
        """
        :return: The report name
        """
        # name = 'name'
        name = ""
        if cls.report_title:
            name = cls.report_title
        elif cls.page_title:
            name = cls.page_title
        return capfirst(name)

    def get_metadata(self, generator):
        metadata = super().get_metadata(generator)
        metadata["report_title"] = self.report_title
        return metadata

    def order_results(self, data):
        """
        order the results based on GET parameter or default_order_by
        :param data: List of Dict to be ordered
        :return: Ordered data
        """
        order_field, asc = OrderByForm(self.request.GET).get_order_by(
            self.default_order_by
        )
        if order_field:
            data = dictsort(data, order_field, asc)
        return data

    def prepare_results_for_printing(self, results):
        """
        Hook before sending the results for teh printing generator
        :param results: dict containing the data, columns, column_names, verbose_data and other keys
        :return: adjusted results
        """
        return results

    def form_invalid(self, form):
        return JsonResponse(form.errors, status=400)

    # @classmethod
    # def get_default_from_date(cls, **kwargs):
    #     return app_settings.RA_DEFAULT_FROM_DATETIME
    #
    # @classmethod
    # def get_default_to_date(cls, **kwargs):
    #     return app_settings.RA_DEFAULT_TO_DATETIME


class TimeSeriesSelectorReportView(UserPassesTestMixin, SlickReportViewBase):
    @staticmethod
    def form_filter_func(fkeys_dict):
        output = {}
        for k, v in fkeys_dict.items():
            if k not in ["owner_id", "polymorphic_ctype_id", "lastmod_user_id"]:
                output[k] = v
        return output

    @classmethod
    def get_form_class(cls):
        """
        As Report Model might be swapped, form_class cant be accurately generated on model load,
        hence this function.
        :return: form_class
        """
        return cls.form_class or report_form_factory(
            cls.get_report_model(),
            crosstab_model=cls.crosstab_model,
            display_compute_reminder=cls.crosstab_compute_reminder,
            fkeys_filter_func=cls.form_filter_func,
            initial=cls.get_form_initial(),
        )
